# -*- coding: utf-8 -*-


import importlib
import tornado.gen
import tornado.concurrent

from dp_tornado.engine.handler import Handler


class SchedulerHandler(Handler):
    @property
    def executor(self):
        return self._executor('worker')

    def _set_response(self, plain):
        setattr(self, '__response', plain)

    def _get_response(self):
        return getattr(self, '__response', None)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self, path=None):
        yield self.get_process(path=path)

        res = self._get_response()

        if res:
            self.finish(res)

    @tornado.concurrent.run_on_executor
    def get_process(self, path):
        if path == 'ping':
            self._set_response('pong')
            return

        try:
            module = importlib.import_module(path)
            runner = getattr(module, 'Scheduler')()
            runner.run()

        except Exception as e:
            self.logging.exception(e)



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper


class SerializationHelper(dpHelper):
    def serialize(self, obj, method='json', beautify=False, raise_exception=False):
        if method == 'json':
            return self.helper.string.serialization.json.stringify(
                obj=obj, beautify=beautify, raise_exception=raise_exception)
        else:
            raise Exception('The specified method is not supported.')

    def deserialize(self, text, method='json', encoding='utf8', raise_exception=False):
        if method == 'json':
            return self.helper.string.serialization.json.parse(
                text=text, encoding=encoding, raise_exception=raise_exception)
        else:
            raise Exception('The specified method is not supported.')



# -*- coding: utf-8 -*-


import re

from dp_tornado.engine.helper import Helper as dpHelper


regex_phone_number = re.compile(
    u"^("
    u"((\+82)(-)\d{1}(-)\d{3,4}(-)\d{4})|"
    u"((\+82)?(-)?\d{3}(-)?\d{4}(-)?\d{4})|"
    u"((\+82)?(-)?\d{3}(-)?\d{3}(-)?\d{4})|"
    u"((\+82)?(-)?\d{2}(-)?\d{4}(-)?\d{4})|"
    u"((\+82)?(-)?\d{2}(-)?\d{3}(-)?\d{4})|"
    u"((\+82)?(-)?\d{4}(-)?\d{4})"
    u")$")


class KoreaHelper(dpHelper):
    def readable_phone_number(self, number, separator='-'):
        number = str(self.helper.numeric.extract_numbers(number))
        l = len(number)

        if l == 8:
            return '%s%s%s' % (number[0:4], separator, number[4:8])
        elif l == 9:
            return '%s%s%s%s%s' % (number[0:2], separator, number[2:5], separator, number[5:9])
        elif l == 10:
            three = ("010", "011", "016", "017", "018", "019",
                     "053", "031", "032", "033", "041", "042",
                     "043", "051", "052", "054", "055", "061",
                     "062", "063", "064", "070", "060", "050")

            if number[0:3] in three:
                return '%s%s%s%s%s' % (number[0:3], separator, number[3:6], separator, number[6:10])
            else:
                return '%s%s%s%s%s' % (number[0:2], separator, number[2:6], separator, number[6:10])
        elif l == 11:
            return '%s%s%s%s%s' % (number[0:3], separator, number[3:7], separator, number[7:11])
        else:
            return number

    def validate_phone_number(self, number):
        return True if re.match(regex_phone_number, number) else False

    def weekday(self, w, short=False, isoweekday=True):
        weekdays = {
            0: '월',
            1: '화',
            2: '수',
            3: '목',
            4: '금',
            5: '토',
            6: '일'
        }

        w = int(w)

        if isoweekday:
            w -= 1

        return '%s요일' % weekdays[w] if not short else weekdays[w]



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

try:
    from wand.image import Image
except ImportError:
    Image = None


class WandHelper(dpHelper):
    @property
    def Image(self):
        return Image

    def load(self, src):
        return Image(filename=src)

    def size(self, src):
        return src.width, src.height

    def crop(self, img, left, top, right, bottom):
        img.crop(left, top, right, bottom)
        return img

    def resize(self, img, width, height, kwargs=None):
        if kwargs is None:
            img.resize(width, height)
        else:
            raise Exception('Not implemented method.')

        return img

    def border(self, img, border, border_color):
        raise Exception('Not implemented method.')

    def radius(self, img, radius, border, border_color):
        raise Exception('Not implemented method.')

    def colorize(self, img, colorize):
        raise Exception('Not implemented method.')

    def save(self, img, ext, dest, kwargs):
        if ext.lower() == 'jpg':
            ext = 'jpeg'

        img.format = ext
        img.save(filename=dest)

        return True

    def iter_seqs(self, img, kwargs):
        yield 0, img



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

try:
    from PIL import Image
except ImportError:
    Image = None

try:
    from PIL import ImageDraw
except ImportError:
    ImageDraw = None

try:
    from PIL import ImageOps
except ImportError:
    ImageOps = None


class PillowHelper(dpHelper):
    @property
    def Image(self):
        return Image.Image

    def load(self, src):
        return Image.open(src)

    def size(self, src):
        return src.width, src.height

    def crop(self, img, left, top, right, bottom):
        return img.crop((left, top, right, bottom))

    def resize(self, img, width, height, kwargs=None):
        if kwargs is None:
            img = img.resize((width, height), Image.ANTIALIAS)
        else:
            background = kwargs['background'] if 'background' in kwargs else None
            mode = kwargs['mode'] if 'mode' in kwargs else None

            if mode == self.helper.io.image.mode.center:
                width_new, height_new = (width, height)
                width_origin, height_origin = img.size

                width_calc = width_new
                height_calc = height_origin * width_calc / width_origin

                if height_calc > height_new:
                    height_calc = height_new
                    width_calc = width_origin * height_calc / height_origin

                img = img.resize((int(width_calc), int(height_calc)), Image.ANTIALIAS)
                img = self.helper.io.image.radius(img, **kwargs)

                img = ImageOps.expand(
                    img, border=(int((width - width_calc) / 2), int((height_new - height_calc) / 2)), fill=background)

                radius = int(kwargs['radius'] or 0) if 'radius' in kwargs else None

                if radius:
                    img_empty = Image.new('RGB', (width, height_new), background)
                    img_empty.paste(img, (0, 0), img)
                    img = img_empty

                img.__dict__['__radius_processed__'] = True

        return img

    def border(self, img, border, border_color):
        if img.mode == 'P':
            img = img.convert('RGBA')

        img = ImageOps.expand(img, border=(border, border), fill=border_color)

        return img

    def radius(self, img, radius, border, border_color):
        img = img.convert('RGBA')
        img.putalpha(self._rounded_mask(img.size, min(radius, *img.size)))

        # Border with radius
        if border:
            img_o = img
            bordered_size = (img.size[0] + (border * 2), img.size[1] + (border * 2))
            img = Image.new('RGBA', bordered_size, border_color)
            img.putalpha(self._rounded_mask(bordered_size, min(radius, *bordered_size)))
            img.paste(img_o, (border, border), img_o)

        return img

    def colorize(self, img, colorize):
        img = img.convert('RGBA')
        r, g, b, a = img.split()
        gray = ImageOps.grayscale(img)
        result = ImageOps.colorize(gray, colorize, (255, 255, 255, 0))
        result.putalpha(a)
        img = result

        return img

    def save(self, img, ext, dest, kwargs):
        # PNG : 1, L, P, RGB, RGBA
        # GIF : L, P
        # JPEG : L, RGB, CMYK

        # Invalid format specified
        if ext.lower() in ('jpg', 'jpeg') and img.mode == 'P':  # JPEG does not support P mode
            img = img.convert('RGBA')

        if ext.lower() == 'jpg':
            ext = 'jpeg'

        args = {
            'format': ext,
            'quality': 100
        }

        if self._animatable(img, kwargs):
            transparency = img.info['transparency'] if 'transparency' in img.info else None
            background = img.info['background'] if 'background' in img.info else None

            # PILLOW Issue, #1592 (github)

            seqs = []

            for i, im in self._iter_seqs(img, kwargs):
                im = im.convert('RGBA')

                if transparency is not None:
                    im.info['transparency'] = transparency
                if background is not None:
                    im.info['background'] = background

                seqs.append(im)

            img = seqs[0]
            seqs.remove(img)

            if transparency is not None:
                args['transparency'] = transparency
            if background is not None:
                args['background'] = background

            args['append_images'] = seqs
            args['save_all'] = True
            args['format'] = 'GIF'

        img.save(dest, **args)

        return True

    def iter_seqs(self, img, kwargs):
        if '__frames__' in img.__dict__ and img.__dict__['__frames__']:
            yield 0, img

            for i in range(1, min(len(img.__dict__['__frames__']), self._iterable_frames(kwargs)) - 1):
                yield i, img.__dict__['__frames__'][i]

        elif not self._animatable(img, kwargs):
            yield 0, img

        else:
            for i in range(min(img.n_frames, self._iterable_frames(kwargs))):
                img.seek(i)
                yield i, img

    def _iterable_frames(self, kwargs):
        if kwargs['frame'] is True or not self.helper.misc.type.check.numeric(kwargs['frame']):
            return 10**7
        else:
            return kwargs['frame']

    def _animatable(self, img, kwargs):
        if '__frames__' in img.__dict__ and img.__dict__['__frames__']:
            return True

        if 'frame' not in kwargs or not kwargs['frame']:
            return False

        if not img:
            return False

        if not img.format or img.format.upper() != 'GIF':
            return False

        if not img.mode or img.mode.upper() != 'P':
            return False

        if not getattr(img, 'n_frames', None) or img.n_frames <= 1:
            return False

        return True

    def _rounded_mask(self, size, radius, factor=2):
        width, height = size
        radius = min(radius, *size)

        mask = Image.new("L", (width * factor, height * factor))
        mask_draw = ImageDraw.Draw(mask)
        mask_draw.ellipse((0, 0, radius * factor, radius * factor), fill='#FFFFFF')

        flip = mask.transpose(Image.FLIP_LEFT_RIGHT)
        mask.paste(flip, (0, 0), flip)
        mask_draw.rectangle((radius * factor / 2, 0, (width - (radius / 2)) * factor, radius * factor), fill='#FFFFFF')
        flip = mask.transpose(Image.FLIP_TOP_BOTTOM)
        mask.paste(flip, (0, 0), flip)
        mask_draw.rectangle((0, radius * factor / 2, width * factor, (height - (radius / 2)) * factor), fill='#FFFFFF')
        mask = mask.resize((width, height), Image.ANTIALIAS)

        return mask




# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class DeleteHelper(dpHelper):
    def raw(self, url, **kwargs):
        return self.helper.web.http.request(req_type='delete', res_type='raw', url=url, **kwargs)

    def json(self, url, **kwargs):
        return self.helper.web.http.request(req_type='delete', res_type='json', url=url, **kwargs)

    def text(self, url, **kwargs):
        return self.helper.web.http.request(req_type='delete', res_type='text', url=url, **kwargs)

    def html(self, url, **kwargs):
        return self.helper.web.http.request(req_type='delete', res_type='html', url=url, **kwargs)



# -*- coding: utf-8 -*-

"""Schema module provides auto-generate relational database model as object based.

File/Class Invoke rules
-----------------------
* */schema/__init__.py*, **DO NOT IMPLEMENT ANY CODE IN THIS FILE**
* */schema/blog/__init__.py*, ``BlogSchema`` > **schema.blog**
* */schema/blog/articles.py*, ``ArticlesSchema`` > **schema.blog.articles**
* */schema/blog/comments.py*, ``CommentsSchema`` > **schema.blog.comments**

Table migration(create or alter) will be executed by invoking ``.migrate()`` method manually.
For example, you can migrate ``comments`` table with invoking ``schema.blog.comments.migrate()`` method.

There are 2 types of schema. One is **__dsn__ specifier** and another one is **table describer**.
``__init__.py`` file of each directory must be **__dsn__ specifier**.
and child ``.py`` files are must be **table describer**.

Here is a **__dsn__ specifier** example **(/schema/blog/__init__.py)**:

.. testcode::

    from dp_tornado.engine.schema import Schema as dpSchema


    class BlogSchema(dpSchema):
        __dsn__ = 'tests.model_test/drv_mysql_test'


and Here is a **table describer** example **(/schema/blog/comments.py)**:

.. testcode::

    from dp_tornado.engine.schema import Table as dpTable
    from dp_tornado.engine.schema import Schema as dpSchema
    from dp_tornado.engine.schema import Attribute as dpAttribute

    class CommentsSchema(dpSchema):
        __table_name__ = 'comments'

        __engine__ = 'InnoDB'
        __charset__ = 'utf8'

        comment_id = dpAttribute.field(dpAttribute.DataType.BIGINT, ai=True, pk=True, nn=True, un=True, comment='Comment ID')
        article_id = dpAttribute.field(dpAttribute.DataType.BIGINT, nn=True, un=True, comment='Parent Article ID')
        author = dpAttribute.field(dpAttribute.DataType.VARCHAR(32), nn=True, comment='Author')
        comment = dpAttribute.field(dpAttribute.DataType.VARCHAR(128), nn=True, comment='Comment')

        primary_key = dpAttribute.index(dpAttribute.IndexType.PRIMARY, 'comment_id')

        idx_articles_author_and_comment = dpAttribute.index(dpAttribute.IndexType.INDEX, ('author', 'comment'))
        idx_articles_comment = dpAttribute.index(dpAttribute.IndexType.INDEX, 'comment')

        fk_comments_article_id = dpAttribute.foreign_key(('article_id', dpSchema.field().blog.articles, 'article_id'))
"""

import inspect

from ..singleton import Singleton as dpSingleton
from ..engine import Engine as dpEngine
from ..loader import Loader as dpLoader
from ..model import Model as dpModel
from ..cache import dpInValueModelConfig


class _ComparableDataType(object):
    name = None
    size = None
    enums = None

    def __eq__(self, other):
        if not other:
            return False

        if self.name == 'ENUM':
            return list(self.enums) == list(other.enums)

        return self.name == other.name and self.size == other.size

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        if self.name == 'ENUM':
            return 'TYPE : %s / ENUMS : %s' % (self.name, list(self.enums))

        return 'TYPE : %s / SIZE : %s' % (self.name, self.size)


class _DataType(object):
    class INT(_ComparableDataType):
        name = 'INT'
        size = 11

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class TINYINT(_ComparableDataType):
        name = 'TINYINT'
        size = 4

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class SMALLINT(_ComparableDataType):
        name = 'SMALLINT'
        size = 6

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class MEDIUMINT(_ComparableDataType):
        name = 'MEDIUMINT'
        size = 9

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class BIGINT(_ComparableDataType):
        name = 'BIGINT'
        size = 20

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class DOUBLE(_ComparableDataType):
        name = 'DOUBLE'
        size = None

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class FLOAT(_ComparableDataType):
        name = 'FLOAT'
        size = None

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class DECIMAL(_ComparableDataType):
        def __init__(self, m=10, d=0):
            self.name = 'DECIMAL'
            self.size = (m, d)

    class CHAR(_ComparableDataType):
        name = 'CHAR'
        size = 1

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class VARCHAR(_ComparableDataType):
        name = 'VARCHAR'
        size = 128

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class TEXT(_ComparableDataType):
        name = 'TEXT'
        size = None

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class TINYTEXT(_ComparableDataType):
        name = 'TINYTEXT'
        size = None

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class MEDIUMTEXT(_ComparableDataType):
        name = 'MEDIUMTEXT'
        size = None

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class LONGTEXT(_ComparableDataType):
        name = 'LONGTEXT'
        size = None

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class ENUM(_ComparableDataType):
        name = 'ENUM'

        def __init__(self, *enums):
            if enums:
                self.enums = enums

    class BLOB(_ComparableDataType):
        name = 'BLOB'
        size = None

        def __init__(self):
            pass

    class LONGBLOB(_ComparableDataType):
        name = 'LONGBLOB'
        size = None

        def __init__(self):
            pass

    class MEDIUMBLOB(_ComparableDataType):
        name = 'MEDIUMBLOB'
        size = None

        def __init__(self):
            pass

    class TINYBLOB(_ComparableDataType):
        name = 'TINYBLOB'
        size = None

        def __init__(self):
            pass

    class DATETIME(_ComparableDataType):
        name = 'DATETIME'
        size = None

        def __init__(self):
            pass

    class DATE(_ComparableDataType):
        name = 'DATE'
        size = None

        def __init__(self):
            pass

    class TIME(_ComparableDataType):
        name = 'TIME'
        size = None

        def __init__(self):
            pass

    class TIMESTAMP(_ComparableDataType):
        name = 'TIMESTAMP'
        size = None

        def __init__(self):
            pass

    class YEAR(_ComparableDataType):
        name = 'YEAR'
        size = 4

        def __init__(self, size=None):
            if size is not None:
                self.size = size

                assert size == 4

    class BINARY(_ComparableDataType):
        name = 'BINARY'
        size = 1

        def __init__(self, size=None):
            if size is not None:
                self.size = size

    class VARBINARY(_ComparableDataType):
        name = 'VARBINARY'
        size = 1

        def __init__(self, size=None):
            if size is not None:
                self.size = size


class _IndexType(object):
    class PRIMARY(object):
        pass

    class UNIQUE(object):
        pass

    class FULLTEXT(object):
        pass

    class INDEX(object):
        pass


class Table(object):
    __migrated = False
    __identifier = None
    __thread_stricted = False

    def _get_driver(self):
        dsn = getattr(self, '__dsn__', None)
        dsn_config = dsn.split('/')

        database = dsn_config[1]
        dsn_config = dsn_config[0]

        try:
            package = dsn_config.split('.')
            dsn_conf = dpEngine().config

            for p in package:
                dsn_conf = dsn_conf.__getattr__(p)

            dsn_conf = dsn_conf.databases.__getattr__(database)

        except AttributeError:
            dsn_conf = None

        driver = getattr(dsn_conf, 'driver', None) if dsn_conf else None

        if not dsn_conf or not isinstance(dsn_conf, object):
            raise Exception('The specified dsn is invalid.')

        if driver.startswith('mysql'):
            driver = 'mysql'
        elif driver.startswith('sqlite'):
            driver = 'sqlite'
        else:
            driver = None

        if not driver:
            raise Exception('The specified driver from dsn is not supported.')

        return dsn, driver

    def _guarantee_permission(self):
        if not Table.__thread_stricted:
            return True

        cache_config = dpInValueModelConfig('sqlite', 'static_url', pure=True)
        cache_key = 'dp:engine:schema:permission'

        got = dpEngine().cache.get(cache_key, dsn_or_conn=cache_config)
        gox = Table.__identifier if got else None

        if got:
            if got == gox:
                return True

            if got != gox:
                return False

        identifier = dpEngine().helper.misc.uuid.v1()
        Table.__identifier = identifier

        dpEngine().cache.set(cache_key, identifier, dsn_or_conn=cache_config)
        got = dpEngine().cache.get(cache_key, dsn_or_conn=cache_config)

        if got == identifier:
            return True

        return False

    def migrate(self):
        if not self._guarantee_permission():
            return None

        return self._migrate(migrate_data=True)

    def migrate_schema(self):
        if not self._guarantee_permission():
            return None

        return self._migrate(migrate_data=False)

    def _migrate(self, migrate_data=True):
        if self.__migrated:
            return None

        self.__migrated = True

        fields = sorted(inspect.getmembers(self, lambda o: isinstance(o, Field)), key=lambda o: o[1].__field_priority__)
        indexes = sorted(inspect.getmembers(self, lambda o: isinstance(o, Index)), key=lambda o: o[1].__field_priority__)
        foreign_keys = sorted(inspect.getmembers(self, lambda o: isinstance(o, ForeignKey)), key=lambda o: o[1].__field_priority__)

        for k, v in indexes:
            if v.index_type == Attribute.IndexType.PRIMARY:
                oo = 0
                for e in v.fields:
                    oo += 1
                    o = getattr(self, e)
                    o.pk = oo
                    o.nn = True

        dsn, driver = self._get_driver()

        if driver == 'mysql':
            from .driver.mysql_driver import MySqlDriver as SchemaDriver
        elif driver == 'sqlite':
            from .driver.sqlite_driver import SqliteDriver as SchemaDriver
        else:
            from .driver import SchemaDriver

        return SchemaDriver.migrate(dsn, self, fields, indexes, foreign_keys, migrate_data=migrate_data)

    def migrate_data(self):
        if not self._guarantee_permission():
            return None

        return self._migrate_data()

    def _migrate_data(self):
        dsn, driver = self._get_driver()

        if driver == 'mysql':
            from .driver.mysql_driver import MySqlDriver as SchemaDriver
        elif driver == 'sqlite':
            from .driver.sqlite_driver import SqliteDriver as SchemaDriver
        else:
            from .driver import SchemaDriver

        return SchemaDriver.migrate_data(dsn, self)

    def __getattribute__(self, item):
        o = super(Table, self).__getattribute__(item)

        if isinstance(o, PirorityData):
            setattr(o, '__table__', self)

        return o

    def __setattr__(self, key, value):
        return super(Table, self).__setattr__(key, value)


__PriorityDataCompareIgnoreKeys__ = ['pk', 'query', 'name', 'm_pk']


class PirorityData(object):
    __field_priority__ = 0

    def __init__(self, **kwargs):
        Field.__field_priority__ += 1
        self.__field_priority__ = Field.__field_priority__

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__temp = kwargs

    def __str__(self):
        return str(self.__temp)

    def __eq__(self, other):
        for k, v in self.__temp.items():
            if k in __PriorityDataCompareIgnoreKeys__:
                continue

            ov = getattr(other, k, None)

            if dpEngine().helper.misc.type.check.string(v):
                if dpEngine().helper.string.cast.unicode(v) != dpEngine().helper.string.cast.unicode(ov):
                    return False
                else:
                    continue

            if ov != v:
                if k == 'nn':
                    ov = getattr(other, 'm_pk', None)

                    if ov == v:
                        continue

                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class Field(PirorityData):
    pass


class Index(PirorityData):
    pass


class ForeignKey(PirorityData):
    pass


class Schema(dpEngine, dpLoader, dpSingleton):
    @staticmethod
    def field():
        return dpEngine().schema


class Attribute(object):
    DataType = _DataType
    IndexType = _IndexType

    @staticmethod
    def field(data_type,
              default=None,
              comment='',
              m_pk=None,
              pk=None,
              uq=None,
              nn=None,
              un=None,
              zf=None,
              ai=None,
              name=None,
              query=None):
        return Field(
            name=name,
            data_type=data_type,
            default=default,
            comment=comment,
            m_pk=m_pk,
            pk=pk,
            uq=uq,
            nn=True if ai else nn,
            un=un,
            zf=zf,
            ai=ai,
            query=query)

    @staticmethod
    def index(index_type, fields, name=None):
        return Index(
            name=name,
            index_type=index_type,
            fields=fields if isinstance(fields, (list, tuple)) else (fields, ))

    @staticmethod
    def foreign_key(fields, on_delete='NO ACTION', on_update='NO ACTION', name=None):
        return ForeignKey(
            name=name,
            on_delete=on_delete,
            on_update=on_update,
            fields=fields if isinstance(fields, (list, tuple)) else (fields, ))




# -*- coding: utf-8 -*-


import logging
import sys

from .singleton import Singleton as dpSingleton
from functools import wraps


class Decorators(dpSingleton):
    @property
    def deprecated(self):
        return DeprecatedDecorator()


class DeprecatedDecorator(object):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, func):
        @wraps(func)
        def wrapped_func(*args, **kwargs):
            if sys.version_info[0] >= 3:
                logging.warning('** This method has been deprecated > %s **' % func.__qualname__)
            else:
                logging.warning('** This method has been deprecated > %s.%s **' % (func.__module__, func.__name__))

            return func(*args, **kwargs)

        return wrapped_func



# -*- coding: utf-8 -*-


import requests.utils

from dp_tornado.engine.helper import Helper as dpHelper

try:
    import urllib.parse as _parse
except:
    import urllib as _parse

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse


class UrlParse(object):
    def __init__(self, request=None, scheme='', netloc='', path='', params=None, query=None, framgment=''):
        if request and 'X-Proxy-Prefix' in request.headers:
            prefix_path = request.headers['X-Proxy-Prefix'].strip()
            prefix_path = prefix_path[:-1] if prefix_path.endswith('/') else prefix_path

            if path.startswith(prefix_path):
                path = path[len(prefix_path):] or '/'

        self.request = request
        self.scheme = scheme
        self.netloc = netloc
        self.path = path
        self.params = params or {}
        self.query = query or {}
        self.fragment = framgment

    def __str__(self):
        return ('scheme=%s, netloc=%s, path=%s, params=%s, query=%s, fragment=%s'
                % (self.scheme, self.netloc, self.path, self.params, self.query, self.fragment))

    def __eq__(self, other):
        if self.scheme != other.scheme:
            return False
        if self.netloc != other.netloc:
            return False
        if self.path != other.path:
            return False
        if self.fragment != other.fragment:
            return False
        if len(self.query) != len(other.query):
            return False
        for k, v in self.query.items():
            if v != other.query[k]:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def build(self):
        url = ''

        if self.scheme:
            url = '%s://' % self.scheme

        if self.netloc:
            url = '%s%s' % (url, self.netloc)

        if self.path:
            url = '%s%s' % (url, self.path)

        if self.query:
            url = '%s?%s' % (url, _parse.urlencode(self.query))

        return url


class UrlHelper(dpHelper):
    def validate(self, e):
        return self.helper.validator.url.validate(e)

    def join(self, *args):
        base = ''

        for e in args:
            base = urlparse.urljoin(base, e)

        return base

    def parse(self, e):
        if self.helper.misc.type.check.string(e):
            uri = self.helper.string.cast.string(e)
            request = None
        else:
            request = e.request

            if hasattr(e, 'request_uri'):
                request_uri = e.request_uri()
            elif hasattr(request, 'uri'):
                request_uri = request.uri
            else:
                raise Exception

            uri = '%s://%s%s' % (request.protocol, request.host, request_uri)

        if self.helper.misc.system.py_version <= 2:
            p = urlparse.urlparse(uri)
            query = dict(urlparse.parse_qsl(p.query, keep_blank_values=True))
        else:
            p = _parse.urlparse(uri)
            query = dict(_parse.parse_qsl(p.query, keep_blank_values=True))

        return UrlParse(request, p.scheme, p.netloc, p.path, p.params, query, p.fragment)

    def build(self, url=None, query=None):
        if self.helper.misc.type.check.string(url):
            uri = self.parse(url)
        elif not hasattr(url, 'build'):
            uri = self.parse('')
        else:
            uri = url

        if query:
            for k, v in query.items():
                uri.query[k] = v

        return uri.build()

    def encode(self, *args, **kwargs):
        return self.quote(*args, **kwargs)

    def decode(self, *args, **kwargs):
        return self.unquote(*args, **kwargs)

    def quote(self, e, safe='', **kwargs):
        if self.helper.misc.system.py_version <= 2:
            return requests.utils.quote(e, safe=safe)
        else:
            return requests.utils.quote(e, safe=safe, **kwargs)

    def unquote(self, e, **kwargs):
        if self.helper.misc.system.py_version <= 2:
            return requests.utils.unquote(e)
        else:
            return requests.utils.unquote(e, **kwargs)



# -*- coding: utf-8 -*-
"""Controller
Accepts input and converts it to commands for the model or view. `<Wikipedia>
<https://en.wikipedia.org/wiki/Model–view–controller>`_

Here is a `foo_bar` controller example:

.. testcode::

    from dp_tornado.engine.controller import Controller

    class FooBarController(Controller):
        def get(self):
            self.finish('done')


Directory/File and URL Mapping rules
------------------------------------
* */controller/__init__.py*, ``StarterController`` > **/**
* */controller/blog/__init__.py*, ``BlogController`` > **/blog**
* */controller/blog/admin/__init__.py*, ``AdminController`` > **/blog/admin**
* */controller/blog/post.py*, ``PostController`` > **/blog/post**
* */controller/blog/view.py*, ``ViewController`` > **/blog/view**
* */controller/foo_bar.py*, ``FooBarController`` > **/foo_bar**


Class/Method and URL Mapping rules
----------------------------------
* */controller/foo.py*, ``def get(self)``: **GET /foo**
* */controller/foo.py*, ``def get(self, a)``: **GET /foo/{path}**
* */controller/foo.py*, ``def get(self, a, b)``: **GET /foo/{path}/{path}**
* */controller/foo.py*, ``def get(self, a, b=None)``: **GET /foo/{path}/{path}** and **GET /foo/{path}**
* */controller/foo.py*, ``def post(self)``: **POST /foo**
* */controller/foo.py*, ``def put(self)``: **PUT /foo**
* */controller/foo.py*, ``def delete(self)``: **DELETE /foo**
* */controller/foo.py*, ``def head(self)``: **HEAD /foo**
* */controller/foo.py*, ``def patch(self)``: **PATCH /foo**
"""


import sys

from .engine import Engine as dpEngine


class NoneValue(object):
    pass

none_value = NoneValue()
py_version = sys.version_info[0]

if py_version >= 3:
    long = int


class Controller(dpEngine):
    """Base class for HTTP request.
    """

    def __init__(self, application, request, **kwargs):
        super(Controller, self).__init__()

        self.prefix = kwargs.get('prefix')
        self.parent = kwargs.get('parent')

        self._write = []
        self._finish = None
        self._render = None
        self._headers = [('Server', 'dp')]

        self.head_requested = False
        self.post_requested = False
        self.get_requested = False
        self.patch_requested = False
        self.delete_requested = False
        self.put_requested = False

    def initialize(self, prefix, parent=None):
        self.prefix = prefix
        self.parent = parent

    @property
    def request(self):
        return self.parent.request

    def get_argument(
            self, name, default=None, strip=True, cast=None, fmt=None, lrange=None, delimiter=None, value=None, **ext):
        value = self.parent.get_argument(name, default, strip) if not value else value

        if value and (cast == 'numeric' or fmt == 'numeric'):  # to Numeric
            value = self.helper.numeric.extract_numbers(value)

        if value and cast == object and fmt in ('yyyymmdd', 'yyyymmddhhiiss'):  # to Datetime Object
            if fmt == 'yyyymmdd':
                value = self.helper.numeric.extract_numbers(value or '')

                if len(value) != 8:
                    return False
            elif fmt == 'yyyymmddhhiiss':
                value = self.helper.numeric.extract_numbers(value or '')

                if len(value) != 14:
                    return False
            else:
                return False
            value = self.helper.datetime.convert(**{fmt: value})
        elif value and cast in self.helper.misc.type.numeric and fmt in ('yyyymmdd', 'yyyymmddhhiiss'):  # to Timestamp
            if fmt == 'yyyymmdd':
                value = self.helper.numeric.extract_numbers(value or '')

                if len(value) != 8:
                    return False
            elif fmt == 'yyyymmddhhiiss':
                value = self.helper.numeric.extract_numbers(value or '')

                if len(value) != 14:
                    return False
            else:
                return False
            value = self.helper.datetime.timestamp.convert(**{fmt: value})

            if cast == self.helper.misc.type.float:
                value = self.helper.numeric.cast.float(value)
            elif cast == self.helper.misc.type.long:
                value = self.helper.numeric.cast.long(value)

        elif value and cast == self.helper.misc.type.int:  # to int
            value = self.helper.numeric.cast.int(value)
        elif value and cast == self.helper.misc.type.long:  # to long
            value = self.helper.numeric.cast.long(value)
        elif value and cast == self.helper.misc.type.float:  # to float
            value = self.helper.numeric.cast.float(value)
        elif value and cast == bool:  # to boolean
            value = str(value).lower()

            if value in ('1', 'yes', 'y', 'true', 't'):
                return True
            elif value in ('0', 'no', 'n', 'false', 'f'):
                return False
            else:
                return -1
        elif value and (cast == 'json' or fmt == 'json'):  # to Json
            return self.helper.string.serialization.deserialize(value)

        if value and fmt == 'url' and not self.helper.validator.url.validate(value):  # URL
            return False
        elif value and fmt == 'email' and not self.helper.validator.email.validate(value):  # Email
            return False
        elif value and fmt == 'email-username' and not self.helper.validator.email.validate('%s@a.com' % value):
            return False
        elif value and fmt == 'email-domain' and not self.helper.validator.email.validate('user@%s' % value):
            return False
        elif value and isinstance(fmt, (list, tuple)) and value not in fmt:
            return False

        if value and fmt == 'html':
            value = self.helper.web.html.validate(value)
        elif value and fmt == 'xss':
            value = self.helper.web.html.strip_xss(value)

        if lrange and len(lrange) == 2:
            if self.helper.misc.type.check.string(value):
                if lrange[1] > len(value) or len(value) > lrange[1]:
                    return False
            elif self.helper.misc.type.check.numeric(value):
                if lrange[1] > value or value > lrange[1]:
                    return False

        # Split
        if value and delimiter and not cast:
            value = value.split(delimiter)

        return value

    def get_user_agent(self, parsed=True, user_agent=None):
        return self.parent.get_user_agent(parsed=parsed, user_agent=user_agent)

    @property
    def remote_ip(self):
        return self.parent.remote_ip

    def set_secure_cookie(self, name, value, expires_days=30, version=2, **kwargs):
        return self.parent.set_secure_cookie(name, value, expires_days=expires_days, version=version, **kwargs)

    def get_secure_cookie(self, name, value=None, max_age_days=31, min_version=None):
        return self.parent.get_secure_cookie(name, value=value, max_age_days=max_age_days, min_version=min_version)

    def secure_cookie(self, name, value=None, **kwargs):
        return self.parent.secure_cookie(name=name, value=value, **kwargs)

    def cookie(self, name, value=None, **kwargs):
        return self.parent.cookie(name=name, value=value, **kwargs)

    def get_sessionid(self):
        return self.parent.get_sessionid()

    def set_sessionid(self, sessionid=None):
        return self.parent.set_sessionid(sessionid=sessionid)

    def get_session_value(self, name, expire_in=None, sessionid=None):
        return self.parent.get_session_value(name, expire_in=expire_in, sessionid=sessionid)

    def get_session_value_ttl(self, name):
        return self.parent.get_session_value_ttl(name)

    def set_session_value(self, name, value, expire_in=none_value):
        if expire_in is none_value:
            expire_in = self.ini.session.expire_in

        return self.parent.set_session_value(name, value, expire_in=expire_in)

    def session(self, name, value=None, expire_in=none_value):
        if expire_in is none_value:
            expire_in = self.ini.session.expire_in

        return self.parent.session(name, value=value, expire_in=expire_in)

    def prefixize(self, url):
        return self.parent.prefixize(url)

    def finish_with_error(self, status_code, message='An error has occurred'):
        return self.parent.finish_with_error(status_code=status_code, message=message)

    def redirect(self, url, prefixize=False, permanent=False, status=None, safe=False):
        if self.parent.headers_written:
            return

        if prefixize:
            url = self.prefixize(url)

        if safe and url and not url.startswith('/'):
            url = '/'

        try:
            self.parent.redirect(url, permanent, status)
        except (AttributeError, TypeError):
            pass

    def request_uri(self, *args, **kwargs):
        return self.parent.request_uri(*args, **kwargs)

    def set_header(self, name, value):
        self._headers.append((name, value))

    def render(self, template_name, kwargs=None):
        self._render = {'t': template_name, 'k': kwargs}

    def render_string(self, template_name, kwargs=None):
        if kwargs:
            return self.view.render_string(self, template_name, kwargs)
        else:
            return self.view.render_string(self, template_name)

    def write(self, chunk):
        self._write.append(chunk)

    def finish(self, chunk=None):
        if isinstance(chunk, (list, tuple, dict)):
            self.set_header('Content-Type', 'application/json; charset=UTF-8')
            chunk = self.helper.string.serialization.serialize(chunk, method='json')

        self._finish = chunk

    def m17n_lang(self, lang=None):
        if lang is None:
            m17n = self.parent.get_cookie('__m17n__')
            return m17n if m17n in self.ini.server.m17n else self.ini.server.m17n[0]

        return self.parent.m17n_lang(lang)



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import json


class JsonHelper(dpHelper):
    def stringify(self, obj, beautify=False, raise_exception=False):
        try:
            if beautify:
                return json.dumps(obj, indent=4, sort_keys=True)
            else:
                return json.dumps(obj, separators=(',', ':'), indent=None, sort_keys=True)

        except Exception as e:
            if raise_exception:
                raise e

            return False

    def parse(self, text, encoding='utf8', raise_exception=False):
        try:
            return json.loads(text, encoding=encoding)

        except Exception as e:
            if raise_exception:
                raise e

            return False



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper
from pytz import timezone as pytz_timezone


class TimezoneHelper(dpHelper):
    def tzinfo_from_zone(self, zone):
        return pytz_timezone(zone)

    def zone_from_tzinfo(self, zone):
        if not zone or not hasattr(zone, 'zone'):
            return None

        return zone.zone

    def tzinfo_from_datetime(self, datetime):
        return datetime.tzinfo

    def zone_from_datetime(self, datetime):
        return self.zone_from_tzinfo(datetime.tzinfo)

    def localize(self, datetime, timezone):
        return timezone.localize(datetime)

    def normalize(self, datetime, timezone):
        if self.helper.misc.type.check.string(timezone):
            timezone = self.tzinfo_from_zone(timezone)

        return timezone.normalize(datetime)



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class CheckHelper(dpHelper):
    def numeric(self, val):
        return True if isinstance(val, self.helper.misc.type.numeric) else False

    def string(self, val):
        return True if isinstance(val, self.helper.misc.type.string) else False

    def array(self, val):
        return True if isinstance(val, self.helper.misc.type.array) else False

    def dict(self, val):
        return True if isinstance(val, self.helper.misc.type.dict) else False



# -*- coding: utf-8 -*-


import logging
import traceback
import threading

from .singleton import Singleton as dpSingleton
from collections import namedtuple

try:
    import Queue as queue
except ImportError:
    import queue


class Logger(dpSingleton):
    def __init__(self, engine=None):
        self.loggers = {}
        self.engine = engine

        self.set_logger(self.default_logger, logging.DEBUG)
        self.set_logger(self.sys_logger, logging.DEBUG)

        self.delegate_queue = None
        self.delegate_handler = None

    @property
    def default_logger_name(self):
        return 'dp_logger'

    @property
    def default_logger(self):
        return self.get_logger(self.default_logger_name)

    @property
    def sys_logger_name(self):
        return 'dp_logger_sys'

    @property
    def sys_logger(self):
        return self.get_logger(self.sys_logger_name)

    def get_logger(self, name):
        if name in self.loggers:
            return self.loggers[name]

        logger = logging.getLogger(name)
        self.loggers[name] = logger

        return logger

    def set_logger(self, logger, level, format_string=None, add_handler=False):
        if format_string is None:
            format_string = '[%(asctime)s][%(levelname)s] %(message)s'

        logger.setLevel(level)
        handler = logging.StreamHandler()
        handler.setLevel(level)

        if add_handler:
            formatter = logging.Formatter(format_string)
            handler.setFormatter(formatter)
            logger.addHandler(handler)

    def set_delegate_handler(self, handler=None):
        self.delegate_handler = handler or self.engine.ini.logging.exception_delegate

    def start_handler(self):
        if self.delegate_handler and not getattr(self, 'delegate', None):
            self.delegate_queue = queue.Queue()
            self.delegate = LoggerDelegate(self)
            self.delegate.start()

    def interrupt(self):
        delegate = getattr(self, 'delegate', None)

        if not delegate:
            return

        self.delegate_interrupt()

    def strip(self, msg, strip=False):
        return msg.strip() if strip else msg

    def exception(self, exception, *args, **kwargs):
        msg = self.strip(str(exception), True)
        tb = traceback.format_exc()

        if self.delegate_handler and self.delegate_queue:
            self.delegate_queue.put((logging.ERROR, msg, tb))

        self.default_logger.exception(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.default_logger.error(self.strip(str(msg), True), *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.default_logger.info(self.strip(msg), *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.default_logger.warning(self.strip(msg), *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        self.default_logger.debug(self.strip(msg), *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        self.default_logger.log(level, self.strip(msg), *args, **kwargs)

    def sys_log(self, msg, *args, **kwargs):
        self.sys_logger.info(self.strip(msg), *args, **kwargs)

    def delegate_interrupt(self):
        if self.delegate_handler:
            self.delegate.interrupted = True
            self.delegate_queue.put(False)

    def set_level(self, name, level):
        self.get_logger(name).setLevel(level)

    @property
    def level(self):
        if hasattr(self, '_level'):
            return getattr(self, '_level')

        levels = namedtuple('LoggingLevel', ('CRITICAL', 'ERROR', 'WARN', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'))

        _level = levels(
            logging.CRITICAL,
            logging.ERROR,
            logging.WARN,
            logging.WARNING,
            logging.INFO,
            logging.DEBUG,
            logging.NOTSET)

        setattr(self, '_level', _level)

        return _level


class LoggerDelegate(threading.Thread):
    def __init__(self, logger):
        self.interrupted = False
        self.logger = logger

        threading.Thread.__init__(self)

    def run(self):
        while not self.interrupted:
            payload = self.logger.delegate_queue.get()

            if not payload:
                break

            self.logger.delegate_handler(*payload)



# -*- coding: utf-8 -*-


class SchemaDriver(object):
    @staticmethod
    def migrate(dsn, table, fields, indexes, foreign_keys):
        raise NotImplementedError

    @staticmethod
    def migrate_data(dsn, table):
        raise NotImplementedError

    @staticmethod
    def _schema_create_table(dsn, table, fields, indexes, foreign_keys):
        raise NotImplementedError



# -*- coding: utf-8 -*-


import boto3

from dp_tornado.engine.helper import Helper as dpHelper


class ItemHelper(dpHelper):
    def put(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            items,
            overwrite_by_pkeys=None):
        items = items if self.helper.misc.type.check.array(items) else (items, )

        return self._batch(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            region_name=region_name,
            table_name=table_name,
            items=items,
            fn_key='Item',
            fn_proc='put_item',
            overwrite_by_pkeys=overwrite_by_pkeys)

    def remove(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            keys,
            overwrite_by_pkeys=None):
        keys = keys if self.helper.misc.type.check.array(keys) else (keys, )

        return self._batch(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            region_name=region_name,
            table_name=table_name,
            items=keys,
            fn_key='Key',
            fn_proc='delete_item',
            overwrite_by_pkeys=overwrite_by_pkeys)

    def get(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            keys):
        multi = True if self.helper.misc.type.check.array(keys) else False
        keys = keys if self.helper.misc.type.check.array(keys) else (keys, )
        output = []

        try:
            table = self.helper.web.aws.dynamodb.table.get(
                access_key_id=access_key_id,
                secret_access_key=secret_access_key,
                region_name=region_name,
                table_name=table_name)

            if not table:
                return False

            for key in keys:
                got = table.get_item(Key=key)

                if not got or 'Item' not in got:
                    continue

                output.append(got['Item'])

            return output if multi else (output[0] if output else [])

        except Exception as e:
            self.logging.exception(e)

            return False

    def query(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            count=False,
            **kwargs):
        table = self.helper.web.aws.dynamodb.table.get(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            region_name=region_name,
            table_name=table_name)

        if count:
            kwargs['Select'] = 'COUNT'

        try:
            result = table.query(**kwargs)
        except Exception as e:
            self.logging.exception(e)

            return False

        if count:
            if 'Count' not in result:
                return False

            return result['Count']

        if not result or 'Items' not in result:
            return []

        return result['Items']

    def scan(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            count=False,
            **kwargs):
        table = self.helper.web.aws.dynamodb.table.get(
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            region_name=region_name,
            table_name=table_name)

        if count:
            kwargs['Select'] = 'COUNT'

        try:
            result = table.scan(**kwargs)
        except Exception as e:
            self.logging.exception(e)

            return False

        if count:
            if 'Count' not in result:
                return False

            return result['Count']

        if not result or 'Items' not in result:
            return []

        return result['Items']

    def _batch(
            self,
            access_key_id,
            secret_access_key,
            region_name,
            table_name,
            items,
            fn_key,
            fn_proc,
            overwrite_by_pkeys=None):
        try:
            table = self.helper.web.aws.dynamodb.table.get(
                access_key_id=access_key_id,
                secret_access_key=secret_access_key,
                region_name=region_name,
                table_name=table_name)

            if not table:
                return False

            with table.batch_writer(overwrite_by_pkeys=overwrite_by_pkeys) as batch:
                fn_proc = getattr(batch, fn_proc)

                for item in items:
                    fn_proc(**{fn_key: item})

            return True

        except Exception as e:
            self.logging.exception(e)

            return False

    def _payloadize(self, payload):
        output = {}

        for k, v in payload.items():
            if v is None:
                output[k] = {'NULL': v}  # A Null data type.
            elif v is True or v is False:
                output[k] = {'BOOL': v}  # A Boolean data type.
            elif self.helper.misc.type.check.numeric(v):
                output[k] = {'N': str(v)}  # A Number data type.
            elif self.helper.misc.type.check.string(v):
                output[k] = {'S': str(v)}  # A String data type.
            elif self.helper.misc.type.check.dict(v):
                output[k] = {'L': self._payloadize(v)}  # A List of attribute values.
            elif self.helper.misc.type.check.array(v):
                tk = 'NS'  # A Number Set data type.

                for e in v:
                    if self.helper.misc.type.check.string(e):
                        tk = 'SS'  # A String Set data type.

                output[k] = {tk: [str(e) for e in v]}
            else:
                output[k] = {'S': str(v)}

        return output



# -*- coding: utf-8 -*-


from dp_tornado.engine.controller import Controller as dpController


class PostController(dpController):
    def get(self):
        params = {
            'request_uri': self.get_argument('request_uri')
        }

        self.render('bbs/post.html', params)

    def post(self):
        validate = self.helper.validator.form.validate(self, {
            'author': {'required': True},
            'title': {'required': True},
            'content': {'required': True}
        }, error_res='code')

        if not validate:
            return

        if not self.model.article.post(**validate):
            return self.finish_with_error(500)

        self.finish({
            'result': True
        })



# -*- coding: utf-8 -*-


from .singleton import Singleton as dpSingleton
from .loader import Loader as dpLoader


class EngineSingleton(dpSingleton):
    @property
    def decorators(self):
        if not hasattr(self, '_decorators'):
            from .decorators import Decorators
            self._decorators = Decorators()

        return self._decorators

    @property
    def ini(self):
        if not hasattr(self, '_ini'):
            from .ini import Initialization as dpIni
            self._ini = dpIni()

        return self._ini

    @property
    def config(self):
        if not hasattr(self, '_config'):
            self._config = dpLoader('config')

        return self._config

    @property
    def schema(self):
        if not hasattr(self, '_schema'):
            self._schema = dpLoader('schema')

        return self._schema

    @property
    def model(self):
        if not hasattr(self, '_model'):
            self._model = dpLoader('model')

        return self._model

    @property
    def helper(self):
        if not hasattr(self, '_helper'):
            self._helper = dpLoader('helper')

        return self._helper

    @property
    def view(self):
        if not hasattr(self, '_view'):
            from .view import View as dpView
            self._view = dpView()

        return self._view

    @property
    def m17n(self):
        if not hasattr(self, '_m17n'):
            from .m17n import M17n as dpM17n
            self._m17n = dpM17n()

        return self._m17n

    @property
    def cache(self):
        if not hasattr(self, '_cache'):
            from .cache import Cache as dpCache
            self._cache = dpCache()

        return self._cache

    @property
    def logger(self):
        if not hasattr(self, '_logger'):
            from .logger import Logger as dpLogger
            self._logger = dpLogger(self)

        return self._logger

    @property
    def vars(self):
        if not hasattr(self, '_vars'):
            from .variable import Variable as dpVariable
            self._vars = dpVariable(self)

        return self._vars

    def executor(self, identifier):
        if not hasattr(self, '_executor_%s' % identifier):

            if self.ini.server.max_worker is None:
                return None

            import tornado.concurrent

            if self.ini.server.max_worker:
                setattr(self,
                        '_executor_%s' % identifier,
                        tornado.concurrent.futures.ThreadPoolExecutor(self.ini.server.max_worker))
            else:
                setattr(self, '_executor_%s' % identifier, None)

        return getattr(self, '_executor_%s' % identifier)


class Engine(object):
    decorators = EngineSingleton().decorators

    @property
    def ini(self):
        return EngineSingleton().ini

    @property
    def options(self):
        return EngineSingleton().options

    @property
    def config(self):
        return EngineSingleton().config

    @property
    def schema(self):
        return EngineSingleton().schema

    @property
    def model(self):
        return EngineSingleton().model

    @property
    def helper(self):
        return EngineSingleton().helper

    @property
    def view(self):
        return EngineSingleton().view

    @property
    def m17n(self):
        return EngineSingleton().m17n

    @property
    def cache(self):
        return EngineSingleton().cache

    @property
    def logger(self):
        return EngineSingleton().logger

    @property
    def logging(self):
        return self.logger

    @property
    def vars(self):
        return EngineSingleton().vars

    def _executor(self, identifier=None):
        return EngineSingleton().executor(identifier)



# -*- coding: utf-8 -*-


import logging

from dp_tornado.engine.schema.driver import SchemaDriver as dpSchemaDriver
from dp_tornado.engine.model import Model as dpModel
from dp_tornado.engine.schema import Attribute as dpAttribute


class SqliteDriver(dpSchemaDriver):
    @staticmethod
    def migrate(dsn, table, fields, indexes, foreign_keys, migrate_data=True):
        succeed = True

        try:
            exists = dpModel().row("""
                SELECT
                    *
                FROM
                    `sqlite_master`
                WHERE
                    `type` = 'table' AND `tbl_name` = :table_name
            """, {'table_name': table.__table_name__}, dsn)

            if not exists:
                SqliteDriver._schema_create_table(
                    dsn=dsn, table=table, fields=fields, indexes=indexes, foreign_keys=foreign_keys)
            else:
                SqliteDriver._schema_alter_table(
                    dsn=dsn, table=table, fields=fields, indexes=indexes, foreign_keys=foreign_keys, exists=exists)

            for e in indexes:
                fields = ','.join(['`%s`' % ee for ee in e[1].fields])
                if e[1].index_type == dpAttribute.IndexType.UNIQUE:
                    tp = 'UNIQUE'
                else:
                    tp = ''

                sql = """
                    CREATE %s INDEX
                        IF NOT EXISTS `%s` ON `%s` (%s)""" % (tp, e[0], table.__table_name__, fields)
                dpModel().execute(sql, None, dsn)

            logging.info('Table migration succeed : %s :: %s' % (dsn, table.__table_name__))

        except Exception as e:
            succeed = True

            logging.exception(e)

            logging.error('Table migration failed : %s :: %s' % (dsn, table.__table_name__))

        if migrate_data:
            return SqliteDriver.migrate_data(dsn, table) and succeed

        return succeed

    @staticmethod
    def migrate_data(dsn, table):
        succed = True
        dummy_data = getattr(table, '__dummy_data__', None)

        if dummy_data:
            dummy_data = dummy_data if isinstance(dummy_data, (tuple, list)) else list(dummy_data)

            proxy = dpModel().begin(dsn)

            try:
                for e in dummy_data:
                    fields = []
                    params = []

                    for k, v in e.items():
                        fields.append(k)
                        params.append(v)

                    proxy.execute("""
                        INSERT OR REPLACE INTO {table_name}
                            ({fields}) VALUES ({params})"""
                                  .replace('{table_name}', table.__table_name__)
                                  .replace('{fields}', ','.join(['`%s`' % ee for ee in fields]))
                                  .replace('{params}', ','.join(['?' for ee in fields])),
                                  params)

                proxy.commit()

                logging.info('Table data insertion succeed : %s :: %s' % (dsn, table.__table_name__))

            except Exception as e:
                succed = False

                proxy.rollback()
                logging.exception(e)

                logging.error('Table data insertion failed : %s :: %s' % (dsn, table.__table_name__))

        return succed

    @staticmethod
    def _schema_create_table(dsn, table, fields, indexes, foreign_keys):
        field_queries = []

        for e in fields:
            sql = SqliteDriver._schema_by_attrs(e)

            if sql:
                field_queries.append(sql)

        field_queries = ',\n'.join(field_queries)

        return dpModel().execute(
            '\n'.join(['CREATE TABLE `%s`(' % table.__table_name__, field_queries, ')']), None, dsn)

    @staticmethod
    def _schema_alter_table(dsn, table, fields, indexes, foreign_keys, exists):
        sql_fields = exists['sql'][exists['sql'].find('(')+1:exists['sql'].rfind(')')]
        sql_fields = ''.join([(e[:e.find('--')] if e.find('--') != -1 else e) for e in sql_fields.split('\n')])
        sql_fields = sql_fields.split(',')

        exist_fields = {}

        for e in sql_fields:
            e = e.strip()
            exist_fields[e[:e.find(' ')]] = True

        for e in fields:
            if (e[1].name or e[0]) in exist_fields:
                continue

            sql = SqliteDriver._schema_by_attrs(e)

            if sql:
                dpModel().execute('ALTER TABLE `%s` ADD COLUMN %s;' % (table.__table_name__, sql), None, dsn)

        return True

    @staticmethod
    def _schema_by_attrs(e):
        field_name = e[1].name or e[0]
        field_type = None
        attr_autoinc = True if e and e[1].ai else False
        attr_notnull = True if e and (e[1].nn or e[1].pk) and not e[1].ai else False

        if e and e[1].data_type.name in ('CHAR', 'VARCHAR', 'TEXT', 'TINYTEXT', 'MEDIUMTEXT', 'LONGTEXT'):
            field_type = 'TEXT'
        elif e and e[1].data_type.name in ('BIGINT', 'INT', 'TINYINT', 'SMALLINT', 'MEDIUMINT'):
            field_type = 'INTEGER'
        elif e and e[1].data_type.name in ('DOUBLE', 'FLOAT'):
            field_type = 'REAL'
        elif e and e[1].data_type.name in ('DECIMAL', ):
            field_type = 'NUMERIC'
        elif e and e[1].data_type.name in ('BLOB', 'LONGBLOB', 'MEDIUMBLOB', 'TINYBLOB', 'BINARY', 'VARBINARY'):
            field_type = 'BLOB'

        if not field_type:
            return None

        attrs = [
            field_name,
            field_type,
            'NOT NULL' if attr_notnull else '',
            'PRIMARY KEY AUTOINCREMENT' if attr_autoinc else ''
        ]

        return ' '.join(attrs).strip()



# -*- coding: utf-8 -*-


from dp_tornado.engine.schema import Table as dpTable
from dp_tornado.engine.schema import Attribute as dpAttribute


class ArticlesSchema(dpTable):
    __table_name__ = 'articles'

    article_id = dpAttribute.field(dpAttribute.DataType.INT, ai=True, pk=True, nn=True, un=True, comment='Article ID')
    title = dpAttribute.field(dpAttribute.DataType.VARCHAR(128), nn=True, comment='Title')
    content = dpAttribute.field(dpAttribute.DataType.LONGTEXT, nn=True, comment='Content')
    author = dpAttribute.field(dpAttribute.DataType.VARCHAR(64), comment='Author')
    signdate = dpAttribute.field(dpAttribute.DataType.INT, comment='Signdate')

    idx_articles_title = dpAttribute.index(dpAttribute.IndexType.INDEX, 'title')
    idx_articles_author = dpAttribute.index(dpAttribute.IndexType.INDEX, 'author')
    idx_articles_signdate = dpAttribute.index(dpAttribute.IndexType.INDEX, 'signdate')

    __dummy_data__ = [
        {'article_id': 1, 'title': 'Helloworld!', 'content': 'Hello!', 'author': 'James', 'signdate': 1478772800},
        {'article_id': 2, 'title': 'Hi!', 'content': 'Hi! Nice to meet you!', 'author': 'Amber', 'signdate': 1478782800},
        {'article_id': 3, 'title': 'Nice to meet you!', 'content': 'Hello. Nice to meet you!', 'author': 'Alice', 'signdate': 1478792800},
        {'article_id': 4, 'title': 'Good morning!', 'content': 'Hi! Good morning!', 'author': 'Elsa', 'signdate': 1478802800},
        {'article_id': 5, 'title': 'Good afternoon!', 'content': 'Hi! Good afternoon!', 'author': 'Kevin', 'signdate': 1478812800},
        {'article_id': 6, 'title': 'Good evening!', 'content': 'Hi! Good evening!', 'author': 'Sam', 'signdate': 1478822800},
        {'article_id': 7, 'title': 'Good night!', 'content': 'Hi! Good night!', 'author': 'Thomas', 'signdate': 1478832800},
        {'article_id': 8, 'title': 'What up!', 'content': 'Hi! What up!', 'author': 'Tim', 'signdate': 1478842800},
        {'article_id': 9, 'title': 'Yo!', 'content': 'Hi! Yo!', 'author': 'William', 'signdate': 1478852800},
        {'article_id': 10, 'title': 'Hor are you?', 'content': 'Hi! What up!', 'author': 'Oscar', 'signdate': 1478862800},
        {'article_id': 11, 'title': 'Good bye!', 'content': 'Hi! Good bye!', 'author': 'Walter', 'signdate': 1478872800},
        {'article_id': 12, 'title': 'See you again!', 'content': 'Hi! See you again!', 'author': 'Jackson', 'signdate': 1478882800},
        {'article_id': 13, 'title': 'Hello!', 'content': 'Hi! What up!', 'author': 'Mike', 'signdate': 1478892800}
    ]



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class RandomHelper(dpHelper):
    def string(self, length):
        return ''.join(self.helper.misc.random.sample(self.helper.string.ascii_letters, length))

    def numeric(self, length):
        return ''.join(self.helper.misc.random.sample(self.helper.string.digits, length))



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import datetime as abs_datetime


class DateHelper(dpHelper):
    def now(self, timezone=None):
        return self.from_datetime(datetime=self.helper.datetime.now(timezone=timezone))

    def from_datetime(self, datetime, timezone=None):
        if not timezone:
            return abs_datetime.datetime(
                year=datetime.year,
                month=datetime.month,
                day=datetime.day,
                tzinfo=datetime.tzinfo)

        if not datetime.tzinfo:
            datetime = self.helper.datetime.timezone.localize(datetime=datetime, timezone=timezone)
        else:
            datetime = self.helper.datetime.timezone.normalize(datetime=datetime, timezone=timezone)

        return self.from_datetime(datetime=datetime)

    def from_timestamp(self, timestamp, timezone=None, ms=False):
        datetime = self.helper.datetime.from_timestamp(timestamp=timestamp, timezone=timezone, ms=ms)
        return self.from_datetime(datetime=datetime)

    def convert(
            self,
            auto=None,
            datetime=None,
            timezone=None,
            timestamp=None,
            yyyymmdd=None,
            yyyymmddhhiiss=None,
            ms=False):
        datetime = self.helper.datetime.convert(
            auto=auto,
            datetime=datetime,
            timezone=timezone,
            timestamp=timestamp,
            yyyymmdd=yyyymmdd,
            yyyymmddhhiiss=yyyymmddhhiiss,
            ms=ms)
        return self.from_datetime(datetime=datetime)

    def year(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False):
        return self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms).year

    def month(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False):
        return self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms).month

    def day(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False):
        return self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms).day

    def weekday(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False, isoweekday=True):
        datetime = self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms)
        return datetime.isoweekday() if isoweekday else datetime.weekday()

    def tuple(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False):
        datetime = self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms)
        time_set = [datetime.year, datetime.month, datetime.day]

        if datetime.tzinfo:
            time_set.append(self.helper.datetime.timezone.zone_from_tzinfo(datetime.tzinfo))

        return time_set

    def yyyymmdd(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False, concat=''):
        datetime = self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms)
        return '%04d%s%02d%s%02d' % (datetime.year, concat, datetime.month, concat, datetime.day)

    def mmdd(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False, concat=''):
        datetime = self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms)
        return '%02d%s%02d' % (datetime.month, concat, datetime.day)



# -*- coding: utf-8 -*-


import os

from dp_tornado import Bootstrap


def run(as_cli=False):
    kwargs = {
        'application_path': os.path.join(os.path.dirname(os.path.realpath(__file__))),
        'as_cli': as_cli
    }

    bootstrap = Bootstrap()
    bootstrap.run(**kwargs)


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-


import tornado.escape


def trim(c, text):
    return text.strip()


def nl2br(c, text, escape=True, break_tag='<br />'):
    if not text:
        return ''

    text = tornado.escape.xhtml_escape(text) if escape else text
    return text.replace('\r\n', break_tag).replace('\r', break_tag).replace('\n', break_tag)


def truncate(c, text, length, ellipsis='..'):
    if len(text) > length:
        return '%s%s' % (text[0:length], ellipsis)

    return text


def yyyymmdd(c, datetime=None, timestamp=None, ms=False, concat='.'):
    return c.helper.datetime.date.yyyymmdd(datetime=datetime, timestamp=timestamp, ms=ms, concat=concat)


def mmdd(c, datetime=None, timestamp=None, ms=False, concat='.'):
    return c.helper.datetime.date.mmdd(datetime=datetime, timestamp=timestamp, ms=ms, concat=concat)


def hhiiss(c, datetime=None, timestamp=None, ms=False, concat=':'):
    return c.helper.datetime.time.hhiiss(datetime=datetime, timestamp=timestamp, ms=ms, concat=concat)


def hhii(c, datetime=None, timestamp=None, ms=False, concat=':'):
    return c.helper.datetime.time.hhii(datetime=datetime, timestamp=timestamp, ms=ms, concat=concat)


def weekday(c, datetime=None, timestamp=None, ms=False, isoweekday=True):
    return c.helper.datetime.date.weekday(datetime=datetime, timestamp=timestamp, ms=ms, isoweekday=isoweekday)


def request_uri(c, with_queries=True, query_quote=False, s=False, d=' ', p='_'):
    r = ('%s%s' % (d, p)).join(c.request.uri.split('/')).strip() if s else c.request.uri
    r = prefixize(c, r)
    r = r if with_queries else r.split('?')[0:1][0]

    return c.helper.web.url.quote(r) if query_quote else r


def m17n(c, m17n_lang=None):
    if not m17n_lang:
        m17n_lang = c.get_cookie('__m17n__')

    return c.m17n.get(m17n_lang)


def get(c, arg, default=None):
    return c.get_argument(arg, default=default)


def number_format(c, val):
    return c.helper.numeric.number_format(val)


def prefixize(c, static_url, query=None, combine_request_query=False, prefix=None, prefix_alternative=None):
    if combine_request_query:
        uri = c.helper.web.url.parse(c)

        if query and isinstance(query, dict):
            query = dict(uri.query, **query)

        elif not query:
            query = uri.query

    if query and isinstance(query, dict):
        uri = c.helper.web.url.parse(static_url)

        for k in query.keys():
            v = query[k]

            if v is None and k in uri.query:
                del uri.query[k]
            elif v is not None:
                uri.query[k] = v

        static_url = c.helper.web.url.build(uri.path, uri.query)

    if prefix or 'X-Proxy-Prefix' in c.request.headers:
        p = prefix_alternative or prefix or c.request.headers['X-Proxy-Prefix'].strip()
        p = p[:-1] if p.endswith('/') else p

        if static_url.startswith(p):
            return static_url[len(p):] or '/'

        return static_url
    else:
        return static_url



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import string


class StringHelper(dpHelper):
    @property
    def ascii_uppercase(self):
        return string.ascii_uppercase

    @property
    def ascii_lowercase(self):
        return string.ascii_lowercase

    @property
    def ascii_letters(self):
        return string.ascii_letters

    @property
    def punctuation(self):
        return string.punctuation

    @property
    def digits(self):
        return string.digits

    @property
    def printable(self):
        return string.printable

    @property
    def whitespace(self):
        return string.whitespace



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import sys


class SysHelper(dpHelper):
    def insert(self, i, x):
        return sys.path.insert(i, x)

    def append(self, x):
        return sys.path.append(x)



# -*- coding: utf-8 -*-

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

import os

from .singleton import Singleton as dpSingleton


class IniSection(object):
    def __init__(self, parser, section):
        self.__dict__['_parser'] = parser
        self.__dict__['_section'] = section
        self.__dict__['_options'] = {}

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            pass

        return self.__dict__['_options'][name] if name in self._options else None

    def __setattr__(self, key, value):
        self.__dict__['_options'][key] = value

    def get(self, key, default=None):
        if key in self.__dict__['_options']:
            return self.__dict__['_options'][key]

        got = self.__dict__['_parser'].get(section=self._section, key=key, default=default)
        self.__setattr__(key, got)

        return got

    def set(self, key, val):
        self.__setattr__(key, val)


class IniParser(object):
    def __init__(self, parser):
        self._parser = parser

    def get(self, section, key, default=None):
        try:
            if not self._parser:
                got = False
            else:
                got = self._parser.get(section, key)

            if default is True or default is False:
                return True if got == '1' else False

            elif isinstance(default, str):
                return str(got)

            elif isinstance(default, int):
                return int(got)

            else:
                return got

        except (configparser.NoSectionError, configparser.NoOptionError):
            return default


class Initialization(dpSingleton):
    def __init__(self):
        self.__dict__['_sections'] = {}
        self.__dict__['_parser'] = None

    @property
    def _parser(self):
        if self.__dict__['_parser']:
            return self.__dict__['_parser']

        application_path = os.getenv('DP_APPLICATION_PATH')
        ini_file = os.getenv('DP_APPLICATION_INI')

        if not application_path or not ini_file:
            parser = None

        else:
            parser = configparser.RawConfigParser()
            parser.read(os.path.join(application_path, ini_file))

        ini_parser = IniParser(parser)
        self.__dict__['_parser'] = ini_parser

        return ini_parser

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            pass

        section = self.__dict__['_sections'][name] if name in self.__dict__['_sections'] else None

        if not section:
            section = IniSection(self._parser, name)
            self.__dict__['_sections'][name] = section

        return section



# -*- coding: utf-8 -*-


import logging

from dp_tornado.engine.schema.driver import SchemaDriver as dpSchemaDriver
from dp_tornado.engine.model import Model as dpModel
from dp_tornado.engine.schema import Attribute as dpAttribute


class MySqlDriver(dpSchemaDriver):
    @staticmethod
    def migrate(dsn, table, fields, indexes, foreign_keys, migrate_data=True):
        for k, v in fields:
            if not v.name:
                setattr(v, 'name', k)

        created = False
        succeed = True

        for val in (False, True):
            proxy = dpModel().begin(dsn)

            try:
                exist = MySqlDriver._get_exist(proxy, table)
                create = MySqlDriver.migrate_fields(proxy, table, fields, exist, change=val)

                if not val:
                    created = create

                if not val:
                    MySqlDriver.migrate_indexes(proxy, table, indexes, exist)
                    MySqlDriver.migrate_foreign_keys(proxy, table, foreign_keys, exist)

                proxy.commit()

                if val:
                    logging.info('Table migration succeed : %s :: %s' % (dsn, table.__table_name__))

            except Exception as e:
                succeed = True

                proxy.rollback()
                logging.exception(e)

                logging.error('Table migration failed : %s :: %s' % (dsn, table.__table_name__))

                break

        if MySqlDriver.migrate_priority(dsn, table, fields) is False:
            return False

        if migrate_data:
            return MySqlDriver.migrate_data(dsn, table) and succeed

        return succeed

    @staticmethod
    def migrate_data(dsn, table):
        succed = True
        dummy_data = getattr(table, '__dummy_data__', None)

        if dummy_data:
            dummy_data = dummy_data if isinstance(dummy_data, (tuple, list)) else list(dummy_data)

            proxy = dpModel().begin(dsn)

            try:
                for e in dummy_data:
                    fields = []
                    params = []

                    for k, v in e.items():
                        fields.append(k)
                        params.append(v)

                    proxy.execute("""
                        INSERT INTO {table_name}
                            ({fields}) VALUES ({params})
                                ON DUPLICATE KEY UPDATE
                                    {updates}"""
                                  .replace('{table_name}', table.__table_name__)
                                  .replace('{fields}', ','.join(['`%s`' % ee for ee in fields]))
                                  .replace('{params}', ','.join(['%s' for ee in fields]))
                                  .replace('{updates}', ','.join(['`%s` = VALUES(`%s`)' % (ee, ee) for ee in fields])),
                                  params)

                proxy.commit()

                logging.info('Table data insertion succeed : %s :: %s' % (dsn, table.__table_name__))

            except Exception as e:
                succed = False

                proxy.rollback()
                logging.exception(e)

                logging.error('Table data insertion failed : %s :: %s' % (dsn, table.__table_name__))

        return succed

    @staticmethod
    def migrate_priority(dsn, table, fields):
        succeed = True
        proxy = dpModel().begin(dsn)

        try:
            exist = MySqlDriver._get_exist(proxy, table)
            exist = [(e[0], e[1]) for e in sorted(exist['col'].items(), key=lambda o: o[1].__field_priority__)]

            exist_priority = [e[0] for e in sorted(exist, key=lambda o: o[1].__field_priority__)]
            exist_fields = dict(exist)

            modified = False

            for i in range(len(fields)):
                if fields[i][0] != exist_priority[i]:
                    modified = True

            if modified:
                previous = None
                queries = []

                for k, v in fields:
                    if previous is None:
                        p = 'FIRST'
                    else:
                        p = 'AFTER `%s`' % previous

                    previous = v.name
                    queries.append('CHANGE COLUMN `%s` %s %s' % (v.name, exist_fields[k].query, p))

                proxy.execute('ALTER TABLE `%s`\n%s' % (table.__table_name__,  ',\n'.join(queries)))

            proxy.commit()

            logging.info('Table priority rearrange succeed : %s :: %s' % (dsn, table.__table_name__))

        except Exception as e:
            succeed = False

            proxy.rollback()
            logging.exception(e)

            logging.error('Table priority rearrange failed : %s :: %s' % (dsn, table.__table_name__))

        return succeed

    @staticmethod
    def _get_chars_pcn(strval):
        l = len(strval)

        for i in range(l):
            yield strval[i-1] if i > 0 else None, strval[i], strval[i+1] if i < l - 1 else None

    @staticmethod
    def _get_field_attr(attrs, begin_offset):
        attrs = attrs[begin_offset:]
        end_offset = 0
        str_val = False

        if attrs and attrs[0] == "'":
            str_val = True

        if not str_val:
            for c in attrs:
                if c == ' ':
                    break
                end_offset += 1

        else:
            cont_quot = False
            attrs = attrs[1:]

            for p, c, n in MySqlDriver._get_chars_pcn(attrs):

                if cont_quot:
                    cont_quot = False
                    end_offset += 1
                    continue

                if c == "'" and n == "'":
                    cont_quot = True
                    end_offset += 1
                    continue

                if end_offset > 0 and c == "'":
                    break

                end_offset += 1

        if not end_offset:
            return None, begin_offset

        return attrs[:end_offset], begin_offset + end_offset + 1

    @staticmethod
    def _get_exist(proxy, table):
        table_name = table.__table_name__

        output = {
            'col': {},
            'ix': [],  # Index objects
            'ik': [],  # Index Keys
            'fk': [],  # Foreign Keys
            'pk': []   # Priamry Keys
        }

        if not proxy.scalar('SHOW TABLES LIKE %s', table_name):
            return output

        offset = 0
        create_table = proxy.row(
            'SHOW CREATE TABLE `{table_name}`'.replace('{table_name}', table_name))[1].split('\n')[1:-1]

        for e in create_table:
            attrs = e.strip()

            if attrs.startswith('`'):
                continue

            if attrs.upper().startswith('KEY'):
                keys = [e[1:-1] for e in attrs[attrs.find('(')+1:attrs.rfind(')')].split(',')]
                output['ix'].append(dpAttribute.index(dpAttribute.IndexType.INDEX, keys))
                output['ik'].append(keys)

            elif attrs.upper().startswith('PRIMARY KEY'):
                keys = [e[1:-1] for e in attrs[attrs.find('(') + 1:attrs.rfind(')')].split(',')]
                output['ix'].append(dpAttribute.index(dpAttribute.IndexType.PRIMARY, keys))
                output['ik'].append(keys)
                output['pk'] = keys

            elif attrs.upper().startswith('UNIQUE KEY'):
                keys = [e[1:-1] for e in attrs[attrs.find('(') + 1:attrs.rfind(')')].split(',')]
                output['ix'].append(dpAttribute.index(dpAttribute.IndexType.UNIQUE, keys))
                output['ik'].append(keys)
                output['pk'] = keys

            elif attrs.upper().startswith('FULLTEXT KEY'):
                keys = [e[1:-1] for e in attrs[attrs.find('(') + 1:attrs.rfind(')')].split(',')]
                output['ix'].append(dpAttribute.index(dpAttribute.IndexType.FULLTEXT, keys))
                output['ik'].append(keys)

            elif attrs.upper().startswith('CONSTRAINT'):
                v, i = MySqlDriver._get_field_attr(attrs, 0)
                fk_name, i = MySqlDriver._get_field_attr(attrs, i)
                foreign, i = MySqlDriver._get_field_attr(attrs, i)
                key, i = MySqlDriver._get_field_attr(attrs, i)

                if foreign.upper() != 'FOREIGN' or key.upper() != 'KEY':
                    continue

                source_col, i = MySqlDriver._get_field_attr(attrs, i)
                dummy, i = MySqlDriver._get_field_attr(attrs, i)
                dest_table, i = MySqlDriver._get_field_attr(attrs, i)
                dest_col, i = MySqlDriver._get_field_attr(attrs, i)

                source_col = [e[1:-1] for e in source_col[1:-1].split(',')]
                dest_table = dest_table[1:-1]
                dest_col = [e[1:-1] for e in dest_col[1:-1].split(',')]

                output['fk'].append((source_col, dest_table, dest_col))

        for e in create_table:
            attrs = e.strip()
            attr = attrs

            if attr[-1] == ',':
                attr = attr[:-1]

            if not attrs.startswith('`'):
                break

            offset += 1

            i = attrs.find(' ')
            column_name = attrs[1:i-1]
            key = column_name

            data_type = None
            default = None
            comment = ''
            nn = None
            un = None
            zf = None
            ai = None

            prev = None
            i += 1

            while True:
                v, i = MySqlDriver._get_field_attr(attrs, i)

                if v is None:
                    break

                if v.strip().upper().startswith('INT('):
                    data_type = dpAttribute.DataType.INT(int(v[4:-1]))
                elif v.strip().upper().startswith('TINYINT('):
                    data_type = dpAttribute.DataType.TINYINT(int(v[8:-1]))
                elif v.strip().upper().startswith('SMALLINT('):
                    data_type = dpAttribute.DataType.SMALLINT(int(v[9:-1]))
                elif v.strip().upper().startswith('MEDIUMINT('):
                    data_type = dpAttribute.DataType.MEDIUMINT(int(v[10:-1]))
                elif v.strip().upper().startswith('BIGINT('):
                    data_type = dpAttribute.DataType.BIGINT(int(v[7:-1]))

                elif v.strip().upper().startswith('DOUBLE('):
                    data_type = dpAttribute.DataType.DOUBLE(int(v[7:-1]))
                elif v.strip().upper().startswith('DOUBLE'):
                    data_type = dpAttribute.DataType.DOUBLE
                elif v.strip().upper().startswith('FLOAT('):
                    data_type = dpAttribute.DataType.FLOAT(int(v[6:-1]))
                elif v.strip().upper().startswith('FLOAT'):
                    data_type = dpAttribute.DataType.FLOAT
                elif v.strip().upper().startswith('DECIMAL('):
                    m, d = v[8:-1].split(',')
                    data_type = dpAttribute.DataType.DECIMAL(int(m), int(d))

                elif v.strip().upper().startswith('CHAR('):
                    data_type = dpAttribute.DataType.CHAR(int(v[5:-1]))
                elif v.strip().upper().startswith('VARCHAR('):
                    data_type = dpAttribute.DataType.VARCHAR(int(v[8:-1]))

                elif v.strip().upper().startswith('TEXT'):
                    data_type = dpAttribute.DataType.TEXT()
                elif v.strip().upper().startswith('TINYTEXT'):
                    data_type = dpAttribute.DataType.TINYTEXT()
                elif v.strip().upper().startswith('MEDIUMTEXT'):
                    data_type = dpAttribute.DataType.MEDIUMTEXT()
                elif v.strip().upper().startswith('LONGTEXT'):
                    data_type = dpAttribute.DataType.LONGTEXT()

                elif v.strip().upper().startswith('BLOB'):
                    data_type = dpAttribute.DataType.BLOB()
                elif v.strip().upper().startswith('LONGBLOB'):
                    data_type = dpAttribute.DataType.LONGBLOB()
                elif v.strip().upper().startswith('MEDIUMBLOB'):
                    data_type = dpAttribute.DataType.MEDIUMBLOB()
                elif v.strip().upper().startswith('TINYBLOB'):
                    data_type = dpAttribute.DataType.TINYBLOB()

                elif v.strip().upper() == 'DATETIME':
                    data_type = dpAttribute.DataType.DATETIME
                elif v.strip().upper() == 'DATE':
                    data_type = dpAttribute.DataType.DATE
                elif v.strip().upper() == 'TIME':
                    data_type = dpAttribute.DataType.TIME
                elif v.strip().upper() == 'TIMESTAMP':
                    data_type = dpAttribute.DataType.TIMESTAMP
                elif v.strip().upper() == 'YEAR':
                    data_type = dpAttribute.DataType.YEAR
                elif v.strip().upper().startswith('YEAR('):
                    data_type = dpAttribute.DataType.YEAR(int(v[5:-1]))

                elif v.strip().upper().startswith('BINARY('):
                    data_type = dpAttribute.DataType.BINARY(int(v[7:-1]))
                elif v.strip().upper().startswith('VARBINARY('):
                    data_type = dpAttribute.DataType.VARBINARY(int(v[10:-1]))

                elif v.strip().upper().startswith('ENUM('):
                    data_type = dpAttribute.DataType.ENUM(*[(e[1:-1] if e[0] == "'" and e[-1] == "'" else e) for e in v[5:-1].split(',')])

                elif v.strip().upper() == 'UNSIGNED':
                    un = True

                elif v.strip().upper() == 'ZEROFILL':
                    zf = True

                elif v.strip().upper() == 'NULL':
                    if prev == 'NOT':
                        nn = True

                elif v.strip().upper() == 'DEFAULT':
                    default, i = MySqlDriver._get_field_attr(attrs, i)
                    default = default[1:-1] if default[0] == "'" and default[-1] == "'" else default
                    default = default if default and default.upper() != 'NULL' else None

                elif v.strip().upper() == 'AUTO_INCREMENT':
                    ai = True

                elif v.strip().upper() == 'COMMENT':
                    comment, i = MySqlDriver._get_field_attr(attrs, i)

                prev = v.upper()

            if comment and comment.endswith('}') and comment.rfind('{') != -1:
                id_idx = comment.rfind('{')
                key = comment[id_idx+1:-1]
                comment = comment[:id_idx-1]

            output['col'][key] = dpAttribute.field(
                data_type=data_type,
                name=column_name,
                default=default,
                nn=nn,
                un=un,
                zf=zf,
                ai=ai,
                comment=comment,
                query=attr)

        return output

    @staticmethod
    def migrate_fields(proxy, table, fields, exist, change=True):
        created = False
        table_name = table.__table_name__

        if not proxy.scalar('SHOW TABLES LIKE %s', table_name):
            fields_query = ',\n'.join(['`%s` %s' % (v.name, MySqlDriver._field_attrs_to_query(k, v, ai=True)) for k, v in fields])
            primary_key = ', '.join(['`%s`' % e[0] for e in sorted([(v.name, v.pk) for k, v in fields if v.pk is not None], key=lambda e: e[1])])
            primary_key = ', PRIMARY KEY (%s)' % primary_key if primary_key else ''

            engine = getattr(table, '__engine__', None)
            charset = getattr(table, '__charset__', None)

            engine = 'ENGINE=%s' % engine if engine else ''
            charset = 'DEFAULT CHARSET=%s' % charset if charset else ''

            proxy.execute("""
                CREATE TABLE `{table_name}` (
                    {fields}
                    {primary_key}
                ) {engine} {charset}"""
                          .replace('{fields}', fields_query)
                          .replace('{primary_key}', primary_key)
                          .replace('{table_name}', table_name)
                          .replace('{engine}', engine)
                          .replace('{charset}', charset))

            return True

        if change:
            exist['col'] = dict(fields)

        for k, v in fields:
            if k in exist['col']:
                if exist['col'][k] != v:
                    proxy.execute("""
                        ALTER TABLE `{table_name}` CHANGE COLUMN `{column_name_b}` `{column_name_a}` {attrs}"""
                                  .replace('{table_name}', table_name)
                                  .replace('{column_name_b}', exist['col'][k].name)
                                  .replace('{column_name_a}', v.name)
                                  .replace('{attrs}', MySqlDriver._field_attrs_to_query(k, v, ai=True)))

                del exist['col'][k]

            else:
                proxy.execute("""ALTER TABLE `{table_name}` ADD COLUMN `{column_name}` {attrs}"""
                              .replace('{table_name}', table_name)
                              .replace('{column_name}', v.name)
                              .replace('{attrs}', MySqlDriver._field_attrs_to_query(k, v, ai=change)))

        for k in exist['col']:
            proxy.execute("""ALTER TABLE `{table_name}` DROP COLUMN `{column_name}`"""
                          .replace('{table_name}', table_name)
                          .replace('{column_name}', exist['col'][k].name))

        return created

    @staticmethod
    def _field_attrs_to_query(key, field, ai=False):
        data_type = field.data_type.name
        unsigned = 'unsigned' if field.un else ''
        default = "DEFAULT '%s'" % field.default if field.default else ''
        comment = "COMMENT '%s {%s}'" % (field.comment, key)
        null = 'NOT NULL' if field.nn or field.ai or field.pk else 'NULL'
        zerofill = 'ZEROFILL' if field.zf else ''
        auto_increment = 'AUTO_INCREMENT' if ai and field.ai else ''

        if getattr(field.data_type, 'size', None) and not zerofill:
            size = getattr(field.data_type, 'size', None)

            if isinstance(size, (tuple, list)):
                data_type = '%s%s' % (data_type, tuple(size))
            else:
                data_type = '%s(%s)' % (data_type, size)

        elif getattr(field.data_type, 'enums', None):
            data_type = '%s%s' % (data_type, getattr(field.data_type, 'enums', None))

        ret = '{data_type} {zerofill} {unsigned} {null} {default} {auto_increment} {comment}' \
            .replace('{data_type}', data_type) \
            .replace('{unsigned}', unsigned) \
            .replace('{null}', null) \
            .replace('{zerofill}', zerofill) \
            .replace('{default}', default) \
            .replace('{auto_increment}', auto_increment) \
            .replace('{comment}', comment)

        return ret.decode('utf8') if dpModel().helper.misc.system.py_version <= 2 else ret

    @staticmethod
    def migrate_indexes(proxy, table, indexes, exist):
        table_name = table.__table_name__

        for k, v in indexes:
            fields = [getattr(table, e).name for e in v.fields]

            if fields in exist['ik']:
                pass
            else:
                index_name = k or '%s_%s' % (table_name, '_'.join(v.fields))
                skip = False

                if v.index_type == dpAttribute.IndexType.PRIMARY:
                    index_type = None
                    skip = True
                elif v.index_type == dpAttribute.IndexType.UNIQUE:
                    index_type = 'UNIQUE INDEX'
                elif v.index_type == dpAttribute.IndexType.FULLTEXT:
                    index_type = 'FULLTEXT INDEX'
                else:
                    index_type = 'INDEX'

                if not skip:
                    proxy.execute('ALTER TABLE {table_name} ADD {index_type} {index_name} ({fields})'
                                  .replace('{table_name}', table_name)
                                  .replace('{index_type}', index_type)
                                  .replace('{index_name}', index_name)
                                  .replace('{fields}', ','.join(['`%s`' % e for e in fields])))

            if v.index_type == dpAttribute.IndexType.PRIMARY:
                if 1 == 2 and exist['pk'] != fields:
                    if exist['pk']:
                        query = 'ALTER TABLE {table_name} DROP PRIMARY KEY, ADD PRIMARY KEY ({fields})'
                    else:
                        query = 'ALTER TABLE {table_name} ADD PRIMARY KEY ({fields})'

                    proxy.execute(query
                                  .replace('{table_name}', table_name)
                                  .replace('{fields}', ','.join(['`%s`' % e for e in fields])))

    @staticmethod
    def migrate_foreign_keys(proxy, table, foreign_keys, exist):
        table_name = table.__table_name__

        for k, v in foreign_keys:
            source_col = v.fields[0]
            source_col = [getattr(table, e).name for e in (source_col if isinstance(source_col, (list, tuple)) else (source_col, ))]
            dest_table_obj = v.fields[1] if v.fields[1] else table
            dest_table = dest_table_obj.__table_name__
            dest_col = v.fields[2]
            dest_col = [getattr(dest_table_obj, e).name for e in (dest_col if isinstance(dest_col, (list, tuple)) else (dest_col, ))]

            fields = (source_col, dest_table, dest_col)

            if fields in exist['fk']:
                pass
            else:
                query = """
                    ALTER TABLE {table_name}
                        ADD CONSTRAINT {fk_name}
                            FOREIGN KEY ({source_col})
                                REFERENCES `{dest_table}` ({dest_col}) ON DELETE {on_delete} ON UPDATE {on_update}"""

                proxy.execute(query
                              .replace('{table_name}', table_name)
                              .replace('{fk_name}', k)
                              .replace('{source_col}', ','.join(['`%s`' % e for e in source_col]))
                              .replace('{dest_table}', dest_table)
                              .replace('{dest_col}', ','.join(['`%s`' % e for e in dest_col]))
                              .replace('{on_delete}', v.on_delete)
                              .replace('{on_update}', v.on_update))



# -*- coding: utf-8 -*-


from dp_tornado.engine.model import Model as dpModel


class LogModel(dpModel):
    def exception_delegate(self, level, msg, traceback):
        print('-------------------------------')
        print(level)
        print(msg)
        print(traceback)



# -*- coding: utf-8 -*-


import requests

from dp_tornado.engine.helper import Helper as dpHelper
from bs4 import BeautifulSoup


class HttpHelper(dpHelper):
    def request(self, req_type, res_type, url, data=None, json=None, raise_exception=False, session=None, **kwargs):
        assert(req_type in ('get', 'post', 'options', 'head', 'delete', 'put', 'patch'))
        assert(res_type in ('raw', 'json', 'text', 'html'))
        assert(not data or isinstance(data, dict))

        if req_type == 'get':
            return self.request_get(
                res_type=res_type, url=url, data=data, raise_exception=raise_exception, session=session, **kwargs)
        elif req_type == 'options':
            return self.request_options(
                res_type=res_type, url=url, data=data, raise_exception=raise_exception, session=session, **kwargs)
        elif req_type == 'head':
            return self.request_head(
                res_type=res_type, url=url, data=data, raise_exception=raise_exception, session=session, **kwargs)
        elif req_type == 'post':
            return self.request_post(
                res_type=res_type, url=url, data=data, json=json,
                raise_exception=raise_exception, session=session, **kwargs)
        elif req_type == 'put':
            return self.request_put(
                res_type=res_type, url=url, data=data, raise_exception=raise_exception, session=session, **kwargs)
        elif req_type == 'patch':
            return self.request_patch(
                res_type=res_type, url=url, raise_exception=raise_exception, session=session, **kwargs)
        elif req_type == 'delete':
            return self.request_delete(
                res_type=res_type, url=url, raise_exception=raise_exception, session=session, **kwargs)

    def request_get(self, res_type, url, data=None, raise_exception=False, session=None, **kwargs):
        kwargs['url'] = url

        if data:
            kwargs['params'] = data

        return self._response(
            res_type=res_type, method='get', raise_exception=raise_exception, session=session, **kwargs)

    def request_options(self, res_type, url, raise_exception=False, session=None, **kwargs):
        kwargs['url'] = url

        return self._response(
            res_type=res_type, method='options', raise_exception=raise_exception, session=session, **kwargs)

    def request_head(self, res_type, url, raise_exception=False, session=None, **kwargs):
        kwargs['url'] = url

        return self._response(
            res_type=res_type, method='head', raise_exception=raise_exception, session=session, **kwargs)

    def request_post(self, res_type, url, data=None, json=None, raise_exception=False, session=None, **kwargs):
        kwargs['url'] = url

        if data:
            kwargs['data'] = data

        if json:
            kwargs['json'] = json

        return self._response(
            res_type=res_type, method='post', raise_exception=raise_exception, session=session, **kwargs)

    def request_put(self, res_type, url, data=None, raise_exception=False, session=None, **kwargs):
        kwargs['url'] = url

        if data:
            kwargs['data'] = data

        return self._response(
            res_type=res_type, method='put', raise_exception=raise_exception, session=session, **kwargs)

    def request_patch(self, res_type, url, data=None, raise_exception=False, session=None, **kwargs):
        kwargs['url'] = url

        if data:
            kwargs['data'] = data

        return self._response(
            res_type=res_type, method='patch', raise_exception=raise_exception, session=session, **kwargs)

    def request_delete(self, res_type, url, raise_exception=False, session=None, **kwargs):
        kwargs['url'] = url

        return self._response(
            res_type=res_type, method='delete', raise_exception=raise_exception, session=session, **kwargs)

    def _response(self, res_type, method, raise_exception=False, session=None, **kwargs):
        session_ = (session if isinstance(session, requests.Session) else requests.Session()) if session else None
        method = getattr(session_ or requests, method)

        try:
            response = method(**kwargs)
            res_code = response.status_code

            if res_type == 'raw':
                res_body = response.content
            elif res_type == 'json':
                res_body = response.json()
            elif res_type == 'text':
                res_body = response.text
            elif res_type == 'html':
                res_body = response.text
                res_body = BeautifulSoup(res_body)
            else:
                return (res_code, False) if not session_ else (session_, res_code, False)

            return (res_code, res_body) if not session_ else (session_, res_code, res_body)

        except Exception as e:
            if raise_exception:
                raise e

            return (False, None) if not session_ else (session_, False, None)



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class IndexingHelper(dpHelper):
    @property
    def partition(self):
        return self.hash

    @property
    def hash(self):
        return 'HASH'

    @property
    def sort(self):
        return self.range

    @property
    def range(self):
        return 'RANGE'



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper


class TimeHelper(dpHelper):
    def hour(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False):
        return self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms).hour

    def minute(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False):
        return self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms).minute

    def second(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False):
        return self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms).second

    def tuple(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False):
        datetime = self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms)
        time_set = [datetime.hour, datetime.minute, datetime.second]

        if ms:
            time_set.append(datetime.microsecond)

        if datetime.tzinfo:
            time_set.append(self.helper.datetime.timezone.zone_from_tzinfo(datetime.tzinfo))

        return time_set

    def hhiiss(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False, concat=''):
        datetime = self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms)
        return '%02d%s%02d%s%02d' % (datetime.hour, concat, datetime.minute, concat, datetime.second)

    def hhii(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False, concat=''):
        datetime = self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms)
        return '%02d%s%02d' % (datetime.hour, concat, datetime.minute)

    def convert(self,
                auto=None,
                datetime=None,
                timezone=None,
                timestamp=None,
                yyyymmdd=None,
                yyyymmddhhiiss=None,
                ms=False):
        return self.helper.datetime.convert(
            auto=auto,
            datetime=datetime,
            timezone=timezone,
            timestamp=timestamp,
            yyyymmdd=yyyymmdd,
            yyyymmddhhiiss=yyyymmddhhiiss,
            ms=ms)



# -*- coding: utf-8 -*-


from tornado.web import StaticFileHandler


class StaticHandler(StaticFileHandler):
    def write_error(self, status_code, **kwargs):
        error = kwargs.get('exc_info', None)
        error = error[1] if error else None
        reason = error.reason if error and getattr(error, 'reason', None) else 'An error has occurred'

        self.set_status(status_code)

        if status_code == 404:
            return self.render('system/http/pp_404.html')

        return self.render('system/http/pp_5xx.html')



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class CsrfHelper(dpHelper):
    _strings = None

    def generate_token(self, controller, token_length=32, token_expire_in=3600*12):
        if CsrfHelper._strings is None:
            CsrfHelper._strings = self.helper.string.digits + \
                                  self.helper.string.punctuation + \
                                  self.helper.string.ascii_letters

        token = ''.join(self.helper.misc.random.sample(CsrfHelper._strings, token_length))
        controller.session('csrf:%s' % token, 'yes', expire_in=token_expire_in)

        return token

    def verify_token(self, controller, key='csrf', value=None):
        token = value or controller.get_argument(key)

        if not token:
            return False

        verified = controller.session('csrf:%s' % token)

        if verified:
            controller.session('csrf:%s' % token, value='')

        return True if verified else False



# -*- coding: utf-8 -*-
"""Here is a `foo_bar` helper example:

.. testcode::

    from dp_tornado.engine.helper import Helper as dpHelper

    class FooHelper(dpHelper):
        def func1(self):
            \"""
            assert self.helper.foo.func1(10, 20) == None
            \"""
            return None

        def func2(self, a):
            \"""
            assert self.helper.foo.func2(10) == 10
            \"""
            return a

        def func3(self, a, b):
            \"""
            assert self.helper.foo.func3(10, 20) == 30
            \"""
            return a + b


File/Class Invoke rules
-----------------------
* */helper/__init__.py*, **DO NOT IMPLEMENT ANY CODE IN THIS FILE**
* */helper/blog/__init__.py*, ``BlogHelper`` > **helper.blog**
* */helper/blog/admin/__init__.py*, ``AdminHelper`` > **helper.blog.admin**
* */helper/blog/post.py*, ``PostHelper`` > **helper.blog.post**
* */helper/blog/view.py*, ``ViewHelper`` > **helper.blog.view**
* */helper/foo_bar.py*, ``FooBarHelper`` > **helper.foo_bar**


Method Invoke rules
-------------------
* */helper/foo.py*, ``def func1(self)``: **helper.foo.func1()**
* */helper/foo.py*, ``def func2(self, a)``: **helper.foo.func2(a)**
* */helper/foo.py*, ``def func3(self, a, b)``: **helper.foo.func3(a, b)**
"""


from .singleton import Singleton as dpSingleton
from .engine import Engine as dpEngine
from .loader import Loader as dpLoader


class Helper(dpEngine, dpLoader, dpSingleton):
    pass



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper


class SerializationHelper(dpHelper):
    def serialize(self, obj, method='json', beautify=False, raise_exception=False):
        """Alias of helper.string.serialization.serialize"""

        return self.helper.string.serialization.serialize(
            obj=obj, method=method, beautify=beautify, raise_exception=raise_exception)

    def deserialize(self, text, method='json', encoding='utf8', raise_exception=False):
        """Alias of helper.string.serialization.deserialize"""

        return self.helper.string.serialization.deserialize(
            text, method=method, encoding=encoding, raise_exception=raise_exception)



# -*- coding: utf-8 -*-


from __future__ import absolute_import
from dp_tornado.engine.helper import Helper as dpHelper

import time
import datetime as abs_datetime


class DatetimeHelper(dpHelper):
    def now(self, timezone=None):
        return abs_datetime.datetime.now(tz=timezone)

    def from_datetime(self, datetime, timezone=None):
        if not timezone:
            return datetime

        if not datetime.tzinfo:
            return self.helper.datetime.timezone.localize(datetime=datetime, timezone=timezone)
        else:
            return self.helper.datetime.timezone.normalize(datetime=datetime, timezone=timezone)

    def from_timestamp(self, timestamp, timezone=None, ms=False):
        timestamp = self.helper.numeric.cast.long(timestamp)

        # TZ Info from Zone
        if timezone and self.helper.misc.type.check.string(timezone):
            timezone = self.helper.datetime.timezone.tzinfo_from_zone(timezone)

        if ms:  # with microseconds
            datetime = abs_datetime.datetime.fromtimestamp(timestamp // 1000, tz=timezone)
            datetime = datetime.replace(microsecond=timestamp % 1000 * 1000)
            return datetime
        else:
            return abs_datetime.datetime.fromtimestamp(timestamp, tz=timezone)

    def transform(self, anonymous):
        if isinstance(anonymous, abs_datetime.datetime):
            return {'datetime': anonymous}

        if self.helper.misc.type.check.string(anonymous):
            anonymous_cast = self.helper.numeric.extract_numbers(anonymous)

            if len(anonymous_cast) != len(anonymous):
                raise Exception('The specified value is invalid.')

            if len(anonymous) == 8:  # yyyymmdd
                return {'yyyymmdd': anonymous}
            elif len(anonymous) == 14:  # yyyymmddhhiiss
                return {'yyyymmddhhiiss': anonymous}
            else:
                raise Exception('The specified value is invalid.')

        elif self.helper.misc.type.check.numeric(anonymous):
            if anonymous < 10000000000:  # timestamp
                return {'timestamp': self.helper.numeric.cast.int(anonymous)}
            else:  # timestamp with ms
                return {'timestamp': self.helper.numeric.cast.long(anonymous), 'ms': True}

        else:
            raise Exception('The specified value is invalid.')

    def convert(self,
                auto=None,
                datetime=None,
                timezone=None,
                timestamp=None,
                yyyymmdd=None,
                yyyymmddhhiiss=None,
                ms=False):
        if auto:
            return self.convert(**self.transform(auto))

        if isinstance(datetime, abs_datetime.datetime):
            if timezone:
                datetime = self.from_datetime(datetime=datetime, timezone=timezone)
        elif timestamp:
            if self.helper.misc.type.check.numeric(timestamp):
                datetime = self.from_timestamp(timestamp=timestamp, timezone=timezone, ms=ms)
            else:
                raise Exception('The specified timestamp value is invalid.')
        elif yyyymmdd:
            if len(self.helper.numeric.extract_numbers(str(yyyymmdd))) == 8:
                timestamp = self.helper.datetime.timestamp.mktime(
                    year=int(yyyymmdd[0:4]),
                    month=int(yyyymmdd[4:6]),
                    day=int(yyyymmdd[6:8]),
                    ms=ms)

                return self.convert(timestamp=timestamp, timezone=timezone, ms=ms)
            else:
                raise Exception('The spcified yyyymmdd value is invalid.')
        elif yyyymmddhhiiss:
            if len(self.helper.numeric.extract_numbers(str(yyyymmddhhiiss))) == 14:
                timestamp = self.helper.datetime.timestamp.mktime(
                    year=int(yyyymmddhhiiss[0:4]),
                    month=int(yyyymmddhhiiss[4:6]),
                    day=int(yyyymmddhhiiss[6:8]),
                    hour=int(yyyymmddhhiiss[8:10]),
                    minute=int(yyyymmddhhiiss[10:12]),
                    second=int(yyyymmddhhiiss[12:14]),
                    ms=ms)

                return self.convert(timestamp=timestamp, timezone=timezone, ms=ms)
            else:
                raise Exception('The spcified yyyymmdd value is invalid.')
        elif not datetime:
            datetime = self.now(timezone=timezone)
        else:
            raise Exception('The specified datetime value is invalid.')

        return datetime

    def tuple(self, auto=None, datetime=None, timezone=None, timestamp=None, ms=False):
        datetime = self.convert(auto=auto, datetime=datetime, timezone=timezone, timestamp=timestamp, ms=ms)
        time_set = [datetime.year, datetime.month, datetime.day, datetime.hour, datetime.minute, datetime.second]

        if ms:
            time_set.append(datetime.microsecond)

        if datetime.tzinfo:
            time_set.append(self.helper.datetime.timezone.zone_from_tzinfo(datetime.tzinfo))

        return time_set



# -*- coding: utf-8 -*-


from .engine import Engine as dpEngine
from .engine import EngineSingleton as dpEngineSingleton
from .model import InValueModelConfig as dpInValueModelConfig
from functools import wraps


class CacheDriver(object):
    def __init__(self, pool=None, conn=None, config_dsn=None):
        self._pool = pool
        self._conn = conn
        self._pipeline = None
        self._config_dsn = config_dsn
        self._reference_count = 0

    @property
    def pool(self):
        return self._pool

    @property
    def conn(self):
        return self._conn

    @staticmethod
    def getpool(config_dsn=None, host=None, port=None, database=None, user=None, password=None):
        pass

    def getconn(self):
        pass


class Cache(dpEngine):
    server_startup_at = None
    executed_pure = {}
    executed_pure_cache_config = dpInValueModelConfig(driver='sqlite', database='executed_pure_cache')

    @staticmethod
    def lock_decorator(*args, **kwargs):
        return LockDecorator(*args, **kwargs)

    @staticmethod
    def decorator(*args, **kwargs):
        return Decorator(*args, **kwargs)

    @staticmethod
    def clear(method, *args, **kwargs):
        kwargs[__cache__clear__] = True
        return method(*args, **kwargs)

    @staticmethod
    def renew(method, *args, **kwargs):
        kwargs[__cache__renew__] = True
        return method(*args, **kwargs)

    @property
    def pools(self):
        if '_pools' not in self.__dict__:
            self.__dict__['_pools'] = {}

        return self.__dict__['_pools']

    @property
    def flags(self):
        if '_flags' not in self.__dict__:
            self.__dict__['_flags'] = {}

        return self.__dict__['_flags']

    def _parse_config(self, config_dsn, delegate):
        if isinstance(config_dsn, dpInValueModelConfig):
            delegate['key'] = 'InValueModelConfig.%s_%s' % (config_dsn.driver, config_dsn.database)
            return config_dsn

        delegate['key'] = config_dsn

        if delegate['key'] in self.pools:
            return self.pools[config_dsn]

        config_dsn = config_dsn.split('/')

        if not config_dsn or len(config_dsn) < 2:
            raise Exception('The specified cache dsn is invalid.')

        config = config_dsn[0]
        database = config_dsn[1]

        try:
            package = config.split('.')
            conf = self.config

            for p in package:
                conf = conf.__getattr__(p)

            conf = conf.caches.__getattr__(database)
        except AttributeError:
            conf = None

        return conf

    def _getdriver(self, config_dsn):
        delegate = {}

        conf = self._parse_config(config_dsn, delegate)
        key = delegate['key']

        if not conf:
            raise Exception('Cache configuration initialized failed.')

        if key not in self.pools:
            if conf.driver == 'redis':
                from .driver.redis_driver import RedisCacheDriver as dpRedisCacheDriver

                self.pools[key] = dpRedisCacheDriver.getpool(
                    key,
                    conf.host,
                    conf.port,
                    conf.database,
                    conf.user,
                    conf.password,
                    conf.maxconn)

            elif conf.driver == 'memory':
                from .driver.memory_driver import MemoryCacheDriver as dpMemoryCacheDriver

                self.pools[key] = dpMemoryCacheDriver.getpool(
                    key
                )

            elif conf.driver == 'sqlite':
                from .driver.sqlite_driver import SqliteCacheDriver as dpSqliteCacheDriver

                self.pools[key] = dpSqliteCacheDriver.getpool(
                    conf
                )

            else:
                raise NotImplementedError

            if key not in self.flags:
                self.flags[key] = True

                if hasattr(self.pools[key], 'create_table'):
                    pure = conf.pure

                    if pure:
                        if not Cache.server_startup_at:
                            import tornado.ioloop
                            i = tornado.ioloop.IOLoop.instance()
                            startup_at = getattr(i, 'startup_at', 0)
                            Cache.server_startup_at = startup_at

                        pure_key = '%s-%s' % (Cache.server_startup_at, str(config_dsn))

                        if pure_key in Cache.executed_pure:
                            pure = False

                        elif pure:
                            cached = self.get(pure_key, dsn_or_conn=Cache.executed_pure_cache_config)

                            if cached:
                                pure = False

                            else:
                                Cache.executed_pure[pure_key] = True
                                self.set(pure_key, 1, dsn_or_conn=Cache.executed_pure_cache_config)

                    self.pools[key].create_table(config_dsn, True if pure else False)

        return self.pools[key]

    def getconn(self, config_dsn):
        driver = self._getdriver(config_dsn)

        if not driver:
            raise Exception('Cache pool initialized failed.')

        return driver.getconn()

    def __getattr__(self, item):
        def delegate(*args, **kwargs):
            dsn_or_conn = kwargs['dsn_or_conn'] if 'dsn_or_conn' in kwargs else None
            dsn_or_conn = kwargs['dsn'] if not dsn_or_conn and 'dsn' in kwargs else dsn_or_conn

            if 'dsn_or_conn' in kwargs:
                del kwargs['dsn_or_conn']

            if 'dsn' in kwargs:
                del kwargs['dsn']

            if not dsn_or_conn:
                raise Exception('The `dsn` argument must specify to invoke a method.')

            config_dsn = dsn_or_conn if isinstance(dsn_or_conn, (str, dpInValueModelConfig)) else None
            conn = self.getconn(config_dsn) if config_dsn else dsn_or_conn
            method = getattr(conn, item, None)

            if not method:
                message = 'The `%s`.`%s` method is not implemented.' % (conn.__class__.__name__, item)
                raise NotImplementedError(message)

            return method(*args, **kwargs)

        return delegate


class LockDecorator(object):
    def __init__(self, a=None, b=None, dsn=None, expire_in=None):
        if a and b is None and isinstance(a, int):
            dsn = dsn or None
            expire_in = expire_in or a
        elif a and b:
            dsn = dsn or a
            expire_in = expire_in or b

        if expire_in is None:
            expire_in = 3600*24*365

        self._dsn = dsn
        self._expire_in = expire_in
        self._expire_at = _engine_.helper.datetime.timestamp.now() + expire_in
        self._func_name = None

    def __call__(self, f):
        @wraps(f)
        def wrapped_f(*args, **kwargs):
            fn_args = args[1:]

            cache_key = 'dp:lock:val:%s:%s:%s' % (args[0].__class__, f.__name__, _engine_.ini.server.identifier)
            acquired = _engine_.cache.setnx(cache_key, self._expire_at, dsn_or_conn=self._dsn, expire_in=self._expire_in)

            if not acquired:
                return False

            try:
                output = f(*args, **kwargs)
            except Exception as e:
                _engine_.cache.delete(cache_key, dsn_or_conn=self._dsn)
                raise e

            _engine_.cache.delete(cache_key, dsn_or_conn=self._dsn)
            return output

        self._func_name = f.__name__

        return wrapped_f


class Storage(object):
    pass

_engine_ = dpEngineSingleton()
_cached_ = Storage()
__cache__clear__ = '__cache__clear__'
__cache__renew__ = '__cache__renew__'


class Decorator(object):
    def __init__(self, a=None, b=None, dsn=None, expire_in=None, **kwargs):
        if a and b is None and isinstance(a, int):
            dsn = dsn or None
            expire_in = expire_in or a
        elif a and b:
            dsn = dsn or a
            expire_in = expire_in or b

        if expire_in is None:
            expire_in = 3600*24*365

        self._dsn = dsn if dsn and isinstance(dsn, (list, tuple)) else (dsn, )
        self._expire_in = expire_in
        self._func_name = None
        self._propagation = kwargs['propagation'] if kwargs and 'propagation' in kwargs else False
        self._ignore = kwargs['ignore'] if kwargs and 'ignore' in kwargs else []

        if not isinstance(self._ignore, (list, tuple)):
            self._ignore = (self._ignore, )

    def _cached(self, cache_key):
        for dsn in self._dsn:
            if dsn:
                cached = _engine_.cache.get(cache_key, dsn_or_conn=dsn)
            else:
                if hasattr(_cached_, cache_key):
                    cached = getattr(_cached_, cache_key, None)
                else:
                    cached = None

            cached = _engine_.helper.string.serialization.deserialize(cached) if cached else None

            if cached is False:
                raise Exception('cached value deserialization failed.')

            elif not cached:
                continue

            elif cached['exp'] and _engine_.helper.datetime.timestamp.now() > cached['exp']:  # Value expired
                if not dsn:
                    delattr(_cached_, cache_key)

                continue

            return cached

        return None

    def _clear(self, cache_key):
        for dsn in self._dsn:
            if not dsn:
                if hasattr(_cached_, cache_key):
                    delattr(_cached_, cache_key)
            else:
                _engine_.cache.delete(cache_key, dsn_or_conn=dsn)

        return True

    def _cache(self, cache_key, value):
        payload = {
            'exp': _engine_.helper.datetime.timestamp.now() + self._expire_in if self._expire_in else None,
            'val': value
        }

        serialized = _engine_.helper.string.serialization.serialize(payload, method='json')

        if not serialized:
            raise Exception('cache decorator supported serializable value only.')

        for dsn in self._dsn:
            if not dsn:
                setattr(_cached_, cache_key, serialized)
            else:
                _engine_.cache.set(key=cache_key, val=serialized, expire_in=self._expire_in, dsn_or_conn=dsn)

        return True

    def _identifier_key(self, a, b):
        return 'dp:cache:key:%s:%s' % (a, b)

    def _identifier(self, identifier_key, ident_renew):
        identifier = self._cached(identifier_key) if not ident_renew else None

        if not identifier:
            identifier = _engine_.helper.datetime.timestamp.now(ms=True)
            self._cache(identifier_key, identifier)
        else:
            identifier = identifier['val']

        return identifier

    def __call__(self, f):
        @wraps(f)
        def wrapped_f(*args, **kwargs):
            cache_clear = False
            cache_renew = False
            ident_renew = False

            fn_args = args[1:]

            if __cache__clear__ in kwargs:
                del kwargs[__cache__clear__]
                cache_clear = True

                if not fn_args and not kwargs:
                    ident_renew = True

            if __cache__renew__ in kwargs:
                del kwargs[__cache__renew__]
                cache_renew = True

            fn_kwargs = dict([(k, v) for k, v in kwargs.items() if k not in self._ignore])

            identifier_key = self._identifier_key(args[0].__class__, f.__name__)
            identifier_key = self._identifier(identifier_key, ident_renew)

            cache_key = 'dp:cache:val:%s:%s:%s:%s-%s' \
                        % (args[0].__class__, f.__name__, identifier_key, fn_args, fn_kwargs.items())

            if cache_renew:
                self._clear(cache_key)

                if self._propagation:
                    kwargs[__cache__renew__] = True

            elif cache_clear:
                self._clear(cache_key)

                if not self._propagation:
                    return True
                else:
                    kwargs[__cache__clear__] = True

            cached = self._cached(cache_key)

            if cached and 'val' in cached:
                return cached['val']

            output = f(*args, **kwargs)

            self._cache(cache_key, output)

            return output

        self._func_name = f.__name__

        return wrapped_f



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper

import os


class PathHelper(dpHelper):
    def join(self, path, *paths):
        return os.path.join(path, *paths)

    def split(self, path):
        return os.path.split(path)

    def mkdir(self, path, mode=None):
        if self.is_file(path):
            path = self.dirname(path)

        try:
            kwargs = {}

            if mode:
                kwargs['mode'] = mode

            if not self.is_dir(path):
                os.makedirs(path, **kwargs)

        except Exception as e:
            self.logging.exception(e)
            pass

        return self.is_dir(path)

    def browse(self, path, fullpath=True, recursive=False, appendpath=None, conditions=None):
        output = [os.path.join(path, e) if fullpath else e for e in os.listdir(path)]

        output_recursive = []

        for e in output:
            if recursive and self.is_dir(e if fullpath else os.path.join(path, e)):
                appendpath_ = (self.join(appendpath, e) if appendpath else e) if not fullpath else None
                output_recursive += self.browse(
                    path=os.path.join(path, e),
                    fullpath=fullpath,
                    recursive=recursive,
                    appendpath=appendpath_,
                    conditions=conditions)
            else:
                if self._browse_filter(path, e, conditions):
                    output_recursive.append(self.join(appendpath, e) if appendpath else e)

        return output_recursive

    def _browse_filter(self, path, name, conditions):
        if not conditions:
            return True

        if 'ext' in conditions:
            ext = self.ext(name, dot='').lower()

            if isinstance(conditions['ext'], (list, tuple)):
                if ext not in conditions['ext']:
                    return False
            elif ext != conditions['ext']:
                return False

        return True

    def is_dir(self, path):
        return True if os.path.isdir(path) else False

    def is_file(self, path):
        return True if os.path.isfile(path) else False

    def dirname(self, path):
        return os.path.dirname(path)

    def cwd(self):
        return os.getcwd()

    def ext(self, path, dot='.'):
        ext = os.path.splitext(path)[-1]

        return '%s%s' % (dot or '', ext[1:]) if ext else ''



# -*- coding: utf-8 -*-


from dp_tornado.engine.engine import Engine


class Processor(Engine):
    def __init__(self, timeout=None):
        pass

    def run(self):
        raise NotImplementedError



# -*- coding: utf-8 -*-


import time
import threading
import pytz
import tornado.httpclient

from dp_tornado.engine.engine import Engine as dpEngine

from concurrent.futures import ThreadPoolExecutor
from multiprocessing import Value


try:
    from croniter import croniter
except ImportError:
    croniter = None


class Scheduler(threading.Thread, dpEngine):
    def __init__(self, schedules):
        self.global_interrupt = Value('d', 0)
        self.global_prepared = Value('d', 0)
        self.schedules = []
        self.ts = self.helper.datetime.timestamp.now()
        self.start_time = self.helper.datetime.now()

        if self.ini.scheduler.mode not in ('web', ):
            raise Exception('The specified scheduler mode is invalid.')

        # Replace timezone
        if self.ini.scheduler.timezone:
            tz = pytz.timezone(self.ini.scheduler.timezone)
            self.start_time = self.helper.datetime.now(timezone=tz)

        job = 0

        for e in schedules:
            opts = e[2] if len(e) >= 3 and isinstance(e[2], dict) else {}  # options
            mode = opts['mode'] if 'mode' in opts and opts['mode'] else self.ini.scheduler.mode  # mode
            period = e[0]  # period or crontab config
            command = e[1]

            clone = 1  # clone count
            repeat = 0  # repeat count (0: infinitely)

            if 'clone' in opts:
                clone = opts['clone']

            if 'repeat' in opts:
                repeat = opts['repeat']

            # init. period config
            if self.helper.misc.type.check.string(period):
                period = croniter(period, start_time=self.start_time)  # crontab
                next_exec = period.get_next()
            else:
                period = self.helper.numeric.cast.int(period) or 1
                next_exec = self.ts

            job += clone

            for i in range(clone):
                self.schedules.append({
                    'command': command,
                    'repeat': repeat,
                    'period': period,
                    'next': next_exec,
                    'mode': mode,
                    'ref': 0,
                    'busy': False,
                    'enabled': True
                })

                dt_next = self.helper.datetime.convert(timestamp=next_exec)
                self.logging.info('Scheduler registered %s, first fetching at %s' % (command, dt_next))

        self.pool = ThreadPoolExecutor(min(64, job * 10))
        threading.Thread.__init__(self)

    def interrupt(self):
        self.global_interrupt.value = 1

    @property
    def interrupted(self):
        return True if self.global_interrupt.value else False

    def prepare(self):
        self.global_prepared.value = 1

    @property
    def prepared(self):
        return True if self.global_prepared.value else False

    def run(self):
        if not self.schedules:
            return

        while not self.interrupted and not self.prepared:
            time.sleep(0.1)

        time.sleep(1.5)

        while not self.interrupted:
            ts = self.helper.datetime.timestamp.now()

            for e in self.schedules:
                if not isinstance(e, dict):
                    continue

                if e['enabled'] and ts >= e['next']:
                    # init. period config
                    if self.helper.misc.type.check.numeric(e['period']):
                        e['next'] = ts + e['period']
                    else:
                        e['next'] = e['period'].get_next()  # crontab

                    dt_next = self.helper.datetime.convert(timestamp=e['next'])

                    if e['busy']:
                        self.logging.warning('Scheduler busy skipped %s, next fetching at %s' % (e['command'], dt_next))
                        continue

                    e['ref'] += 1
                    e['busy'] = True

                    try:
                        self.request(e)

                    except Exception as ex:
                        self.logging.exception(ex)

                    if e['ref'] >= e['repeat'] > 0:
                        e['enabled'] = False
                        self.logging.info('Scheduler executed %s, last execution. done.' % e['command'])
                    else:
                        self.logging.info('Scheduler executed %s, next fetching at %s' % (e['command'], dt_next))

            time.sleep(0.2)

    @property
    def request_host(self):
        return 'http://127.0.0.1:%s/dp/scheduler' % self.ini.server.port

    def request(self, payload):
        def _req():
            url = '%s/%s' % (self.request_host, payload['command'])
            code, content = self.helper.web.http.get.text(url)

            if code != 200:
                payload['ref'] -= 1
                payload['enabled'] = True

            payload['busy'] = False

        self.pool.submit(_req)



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper


class ColumnHelper(dpHelper):
    @property
    def string(self):
        return 'S'

    @property
    def binary(self):
        return 'B'

    @property
    def number(self):
        return 'N'



# -*- coding: utf-8 -*-


class Response(Exception):
    def __init__(self, http_status_code=200, message='An error has occurred'):
        self._http_status_code = http_status_code
        self._message = message

    @property
    def http_status_code(self):
        return self._http_status_code

    @property
    def message(self):
        return self._message

    def __str__(self):
        return '%s %s' % (self.http_status_code, self.message)

    def response(self):
        return str(self)



# -*- coding: utf-8 -*-
"""Here is a `foo_bar` config example:

.. testcode::

    from dp_tornado.engine.config import Config


    class FooBarConfig(Config):
        def index(self):
            self.conf.value_str = 'string'
            self.conf.value_int = 10
            self.conf.value_dict = {
                'foo': 100,
                'bar': {
                    'baz': 200,
                    'baf': 300
                }
            }

            self.conf.databases = {
                'drv_sqlite_name': {
                    'driver': 'sqlite'
                },
                'drv_mysql_name': {
                    'driver': 'mysql+cymysql',
                    'database': '',
                    'host': '127.0.0.1',
                    'port': 3306,
                    'user': 'root',
                    'password': '',
                    'pool_size': 1,
                    'charset': 'utf8'
                }
            }

            self.conf.caches = {
                'drv_sqlite_name': {
                    'driver': 'memory',
                    'identifier': 'dp_test_sqlite'
                },
                'drv_redis_name': {
                    'driver': 'redis',
                    'host': '127.0.0.1',
                    'port': 6379,
                    'password': None,
                    'maxconn': 256
                }
            }


File/Class Invoke rules
-----------------------
* */config/__init__.py*, **DO NOT IMPLEMENT ANY CODE IN THIS FILE**
* */config/blog/__init__.py*, ``BlogConfig`` > **config.blog**
* */config/blog/admin/__init__.py*, ``AdminConfig`` > **config.blog.admin**
* */config/blog/post.py*, ``PostConfig`` > **config.blog.post**
* */config/blog/view.py*, ``ViewConfig`` > **config.blog.view**
* */config/foo_bar.py*, ``FooBarConfig`` > **config.foo_bar**


Config Invoke rules
-------------------
* */config/foo_bar.py*, ``self.conf.value_str``: **config.foo_bar.value_str == 'string'**
* */config/foo_bar.py*, ``self.conf.value_int``: **config.foo_bar.value_int == 10**
* */config/foo_bar.py*, ``self.conf.value_dict``: **config.foo_bar.value_dict.foo == 100**
* */config/foo_bar.py*, ``self.conf.value_dict``: **config.foo_bar.value_dict.bar.baz == 200**
* */config/foo_bar.py*, ``self.conf.value_dict``: **config.foo_bar.value_dict.bar.baf == 300**
* */config/foo_bar.py*, ``self.conf.databases``: **'foo_bar/drv_sqlite_name'**
* */config/foo_bar.py*, ``self.conf.databases``: **'foo_bar/drv_mysql_name'**
* */config/foo_bar.py*, ``self.conf.databases``: **'foo_bar/drv_sqlite_name'**
* */config/foo_bar.py*, ``self.conf.databases``: **'foo_bar/drv_redis_name'**
"""


from .singleton import Singleton as dpSingleton
from .namer import Namer as dpNamer
from .loader import Loader as dpLoader
from .engine import EngineSingleton as dpEngineSingleton


engine = dpEngineSingleton()


class Config(dpLoader, dpSingleton):
    conf = dpNamer()

    def __getattr__(self, name):
        return self.__getattr_inner__(name) or self.conf.__getattr__(name)

    @property
    def ini(self):
        return engine.ini



# -*- coding: utf-8 -*-


from dp_tornado.engine.model import Model as dpModel


class ArticleModel(dpModel):
    def index(self):
        self.schema.bbs.articles.migrate()

    def inquiry(self, article_id=None, page=1, rpp=30):
        where_query = '1 = 1'
        params = []

        if article_id:
            where_query = '{w} AND `article_id` = ?'.replace('{w}', where_query)
            params.append(article_id)

        count = self.scalar("""
            SELECT
                COUNT(*)
            FROM
                `articles`
            WHERE
                {w}
        """.replace('{w}', where_query), params or None, 'db/service')

        params.append(rpp)
        params.append(rpp * (page - 1))

        items = self.rows("""
            SELECT
                *
            FROM
                `articles`
            WHERE
                {w}
            ORDER BY
                `signdate` DESC
            LIMIT ? OFFSET ?
        """.replace('{w}', where_query), params, 'db/service')

        if rpp == 1:
            return items[0] if items else None

        return {
            'page': page,
            'rpp': rpp,
            'count': count,
            'rows': items
        }

    def post(self, author, title, content):
        executed = self.execute("""
            INSERT INTO `articles`
                (`author`, `title`, `content`, `signdate`)
                    VALUES (?, ?, ?, ?)
        """, (author, title, content, self.helper.datetime.timestamp.now()), 'db/service')

        return executed.rowcount



# -*- coding: utf-8 -*-


import validators

from dp_tornado.engine.helper import Helper as dpHelper


class UrlHelper(dpHelper):
    def validate(self, url):
        return True if self.helper.misc.type.check.string(url) and validators.url(url) is True else False



# -*- coding: utf-8 -*-


from dp_tornado.engine.helper import Helper as dpHelper

import shutil
import os
import filecmp


class FileHelper(dpHelper):
    def remove(self, files_or_dirs=None, files=None, dirs=None):
        assert not files_or_dirs or not files or not dirs

        files_or_dirs = files_or_dirs or []
        files = files or []
        dirs = dirs or []

        if not isinstance(files_or_dirs, (list, tuple)):
            files_or_dirs = (files_or_dirs, )

        if not isinstance(files, (list, tuple)):
            files = (files, )

        if not isinstance(dirs, (list, tuple)):
            dirs = (dirs, )

        for f in files_or_dirs:
            if os.path.isdir(f):
                dirs.append(f)
            elif os.path.isfile(f):
                files.append(f)

        removed = []

        for f in files:
            if os.path.isfile(f):
                os.remove(f)
                removed.append(f)

        for f in dirs:
            if os.path.isdir(f):
                shutil.rmtree(f)
                removed.append(f)

        return removed

    def write(self, path, content, mode='w'):
        with open(path, mode) as fp:
            fp.write(content)

    def compare(self, *files):
        if not files:
            return None

        def _iter():
            l = len(files)
            for i in range(l-1):
                yield files[i], files[i+1]

        try:
            for f1, f2 in _iter():
                if not filecmp.cmp(f1, f2):
                    return False
        except Exception as e:
            self.logging.exception(e)

            return False

        return True
